"use strict";
this.name	= "escortdeck";
this.author	= "Norby";
this.copyright	= "2015 Norbert Nagy";
this.description= "Buy and hold escort ships on external deck";
this.licence	= "CC BY-NC-SA 4.0";

this.$Debug = false; //verbose log
this.$Derelicts = false; //add derelict ships near the nav buoy of the starting station
this.$EscortDeckBigTraderLaunch = false; //launch Anaconda from Carriers automatically or not
this.$EscortDeckCMaxMass = 800000; //escorts must below this mass on XL deck of Carriers
this.$EscortDeckExtremeMass = 10*this.$EscortDeckCMaxMass; //so much so don't display message
this.$EscortDeckJumpDist = 30; //escorts jump this m farther from the main ship at launch
this.$EscortDeckMassEffect = true; //slower steering and acceleration with more mass on deck
this.$EscortDeckMinMass = 30000; //which ships can tow the deck (30t not small in ShipVersion)
this.$EscortDeckLandingDist = 500; //escorts will be added to the deck within this m if possible
this.$EscortDeckLargeMass = 130000; //max. escort mass, border of large ships in ShipVersion OXP
this.$EscortDeckWMass = 2 * this.$EscortDeckLargeMass; //which ships can tow wide deck (2*large)
this.$EscortDeckXLMass = 400000; //which can tow XL deck (400t mean xl ships in Telescope OXP)

//internal properties, should not touch
this.$EscortDeckEqs = []; //buyable escort ship sizes and masses
this.$EscortDeckEqs["EQ_ESCORTDECK_adder"] = [28.02, 14.9, 51, 16000];
this.$EscortDeckEqs["EQ_ESCORTDECK_asp"] = [64.73, 23.52, 83.69, 75882];
this.$EscortDeckEqs["EQ_ESCORTDECK_gecko"] = [61.48, 14.64, 51, 24670];
this.$EscortDeckEqs["EQ_ESCORTDECK_sidewinder"] = [63.23, 14.43, 44, 25088];
this.$EscortDeckEqs["EQ_ESCORTDECK_viper"] = [50, 18.12, 65.199, 34000];

this.$EscortDeckPad = { //landing pad max. x,y,z sizes and max. mass
			escort:[70,29,95,this.$EscortDeckLargeMass], //heavy escorts: Asp, Arachnid, S8
			mini:[39,15,65,30000], //Adder, Ophidian
			wide:[95,29,95,this.$EscortDeckLargeMass], //wide ship: Krait, Tiger Ray
			xl:[70,29,95,60000], //sizes like heavy escorts but smaller mass, no Asp
			//Carrier specific pads
			center:[65,29,95,130000], //small center pad for Asp, Fer-De-Lance, S8
			main:[90,64,170,this.$EscortDeckCMaxMass], //Anaconda, D.T.T. Atlas
			small:[65,19,66,40000], //aft side pad for Gecko, Ophidian, Sonoran, Viper
			top:[140,35,99,this.$EscortDeckCMaxMass] } //Cobra Mark III, King Cobra

this.$EscortDeckLPos = [[-36,6,-25], [36,6,-25], //landing pad positions: top left-right
			[-36,-6,-25], [36,-6,-25], //bottom left-right
			[-43,0,-25], [43,0,-25]];  //center left-right (Adder only)
//			[34,0,-43], [-34,0,-43],   //aft left, aft right (facing backward) - disabled
this.$EscortDeckLVec = [[0,1,0], [0,1,0], //vectors from landing pad to ship center: up
			[0,-1,0], [0,-1,0], //down
			[-1,0,0], [1,0,0]]; //left, right
//			[0,0,-1], [0,0,-1], //back - disabled

this.$EscortDeckWPos = [[-48,6,-25], [48,6,-25], //Wide deck landing pad positions: top left-right
			[-48,-6,-25], [48,-6,-25], //bottom left-right
			[-54,0,-25], [54,0,-25]];  //center left-right (Adder only)
//			[34,0,-43], [-34,0,-43],   //aft left, aft right (facing backward)
this.$EscortDeckWVec = [[0,1,0], [0,1,0], //vectors from landing pad to ship center: up
			[0,-1,0], [0,-1,0], //down
			[-1,0,0], [1,0,0]]; //left, right
//			[0,0,-1], [0,0,-1], //back

this.$EscortDeckXLPos = [[-8,0,-15], [8,0,-15], //XL deck landing pad positions: left, right
			[-8,0,-44], [8,0,-44], //left, right
			[-8,0,-73], [8,0,-73], //left, right
			[-8,0,-102], [8,0,-102]];  //left, right
this.$EscortDeckXLVec = [[-1,0,0], [1,0,0], //XL deck vectors from landing pad to ship center
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0]]; //left, right

this.$EscortDeckCPos = [[0,15,215], [0,15,135], //Carrier pad positions: top front and center
			[0,15,15], [-67,16,15], [67,16,15], //aft center, left, right
			[-50,15,46], [50,15,46], //main deck left, right
			[-30,0,233], [30,0,233]];  //mini front deck left, right
this.$EscortDeckCSize = [this.$EscortDeckPad.top, //Carrier max. sizes and mass, top front pad
			this.$EscortDeckPad.center, //center: Asp, Fer-De-Lance, S8
			this.$EscortDeckPad.top, //top aft for Cobra Mark III and King Cobra
			this.$EscortDeckPad.small, this.$EscortDeckPad.small, //aft sides
			this.$EscortDeckPad.main, this.$EscortDeckPad.main, //left-right main deck
			this.$EscortDeckPad.mini, this.$EscortDeckPad.mini]; //front pads
this.$EscortDeckCVec = [[0,1,0], [0,1,0], //Carrier vectors from landing pad to ship center: up
			[0,1,0], //up
			[0,1,0], [0,1,0], //up
			[0,0,1], [0,0,1], //forward
			[0,0,0], [0,0,0]]; //center

this.$EscortDeckCXPos = [[0,15,215], [0,15,115],  //Carrier XL deck positions: top front, center
			[-48,16,15], [48,16,15], //aft left, right
			[-11,0,151], [11,0,151], //left, right
			[-11,0,81], [11,0,81], //left, right
			[-30,0,233], [30,0,233]];  //front left, right
this.$EscortDeckCXSize = [this.$EscortDeckPad.top, this.$EscortDeckPad.top, //top front and center
			this.$EscortDeckPad.wide, this.$EscortDeckPad.wide, //top aft wide pads
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort, //4 heavy escorts
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort, //on main decks
			this.$EscortDeckPad.mini, this.$EscortDeckPad.mini]; //front pads
			
this.$EscortDeckCXVec = [[0,1,0], [0,1,0], //Carrier XL deck vectors from landing pad to ship center: up
			[0,1,0], [0,1,0], //up
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0], //left, right
			[0,0,0], [0,0,0]]; //center

/* Carrier deck positions for 10 heavy escorts, removed to be less powerful
this.$EscortDeckCXXPos = [[0,14,220], [0,14,23], //landing pad positions: top front-aft
			[-11,0,170], [11,0,170], //left, right
			[-11,0,135], [11,0,135], //left, right
			[-11,0,100], [11,0,100], //left, right
			[-11,0,65], [11,0,65]];  //left, right
this.$EscortDeckCXXSize = [this.$EscortDeckPad.top, this.$EscortDeckPad.top, //top front and aft
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort, //8 heavy escorts
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort, //on main decks
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort,
			this.$EscortDeckPad.escort, this.$EscortDeckPad.escort];
this.$EscortDeckCXXVec = [[0,1,0], [0,1,0], //vectors from landing pad to ship center: up
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0], //left, right
			[-1,0,0], [1,0,0]]; //left, right
*/
this.$EscortDeckCWS = null; //carriers worldscript if Carriers OXP is installed
this.$EscortDeckDockingTunnel = false; //player in the tunnel
this.$EscortDeckEWS = null; //escortdeck_escort worldscript
this.$EscortDeckFCB = null; //FrameCallBack updating escort ship positions
this.$EscortDeckFCBDelta = 0; //delta counter in FrameCallBack
this.$EscortDeckFirstMSO = true; //first missionScreenOpportunity after docked
this.$EscortDeckLaunch = false; //launch or call back via primeable equipment
this.$EscortDeckLocked = []; //which positions will not launch with the other escorts
this.$EscortDeckMaxs = null; //store maxThrust, maxPitch, maxRoll and maxYaw
this.$EscortDeckPadPos = []; //landing pad positions
this.$EscortDeckPadSize = []; //landing pad sizes
this.$EscortDeckPadVec = []; //vectors from landing pad to ship center
this.$EscortDeckPlayableDataKeys = null;  //playable ships
this.$EscortDeckPWS = null; //escortpack worldscript if escortpack OXP is installed
this.$EscortDeckSelectedPad = -1; //the pad selected via primeable equipment
this.$EscortDeckShip = []; //store the ship objects of escorts
this.$EscortDeckShipData = []; //store the data of escort ships
this.$EscortDeckShipPos = []; //positions of escort ships on board
this.$EscortDeckSitInto = false; //sit into the selected escort when launch it from a Carrier
this.$EscortDeckSitIntoWithMode = false; //allow sit into with mode key when deck is primed
this.$EscortDeckSound = null; //sound of escort landing
this.$EscortDeckSpawnedShips = []; //stack for restoring ship data in FCB after shipSpawned
this.$EscortDeckTWS = null; //telecope worldscript if telecope OXP is installed
this.$EscortDeckToWS = null; //towbar worldscript if towbar OXP is installed
this.$EscortDeckV = null; //visualeffect of the escort deck of the player ship
this.$EscortDeckXL = false; //the current deck is xl or not
this.$EscortDeckZV = 0; //the z offset of deck position from the center of the player ship

//equipment events
this.activated = function() { //launch or call back one or all escorts
	var w = worldScripts.escortdeck;
	w.$EscortDeck_EQActivated(w);
}

this.mode = function() { //select the next pad on escort deck
	var w = worldScripts.escortdeck;
	w.$EscortDeck_EQMode(w);
}

//worldscript events
this.startUp = function() {
	
	this.$EscortDeck_SaveMaxs( player.ship, this );
	if( !player.ship.addCollisionException ) { //before v1.81 need more room
		for( var i = 0; i < this.$EscortDeckLPos.length; i++) 
			this.$EscortDeckLPos[i][2] -= 30;
		for( var i = 0; i < this.$EscortDeckWPos.length; i++)
			this.$EscortDeckWPos[i][2] -= 30;
		for( var i = 0; i < this.$EscortDeckXLPos.length; i++)
			this.$EscortDeckXLPos[i][2] -= 20;
	}
	this.$EscortDeck_SetPads(player.ship, this);
	
	this.$EscortDeckSound = new SoundSource;
	this.$EscortDeckSound.sound = "escortdeck.ogg";
	this.$EscortDeckSound.loop = false;
	this.$EscortDeckSound.repeatCount = 1;
	this.$EscortDeckSound2 = new SoundSource;
	if( worldScripts["BGS-M"] ) this.$EscortDeckSound2.sound = "bgs-c_sell.ogg";
	else this.$EscortDeckSound2.sound = "sell.ogg"; //Oolite default sell sound
	this.$EscortDeckSound2.loop = false;
	this.$EscortDeckSound2.repeatCount = 1;
	this.$EscortDeckSound3 = new SoundSource;
	this.$EscortDeckSound3.sound = "escortdeckhitderelict.ogg";
	this.$EscortDeckSound3.loop = false;
	this.$EscortDeckSound3.repeatCount = 1;

	this.$EscortDeckPlayableDataKeys = Ship.keysForRole( "player" );  //playable ships
	
	var e = worldScripts.escortdeck_escort;
	if( e ) this.$EscortDeckEWS = e; //save escortdeck_escort worldscript
	var c = worldScripts.carriers;
	if( c ) this.$EscortDeckCWS = c; //save Carriers worldscript
	var p = worldScripts.escortpack;
	if( p ) this.$EscortDeckPWS = p; //save EscortPack worldscript
	var h = worldScripts.hudselector;
	if( h && h.$HUDSelectorAddMFD ) h.$HUDSelectorAddMFD(this.name);

	var s = missionVariables.$EscortDeckShipData;
	if( s && s.length > 0 )	this.$EscortDeckShipData = JSON.parse(s);
	var s = missionVariables.$EscortDeckLocked;
	if( s && s.length > 0 )	this.$EscortDeckLocked = JSON.parse(s);

	var t = worldScripts.telescope;
	if( t ) this.$EscortDeckTWS = t; //save Telecope worldscript
	var t = worldScripts.towbar;
	if( t ) {
		this.$EscortDeckToWS = t; //save Towbar worldscript
		//inject code to do not lock on derelicts of escort deck
		eval("t.$EscortDeckToWS_IsDerelictOrig = "+t.$Towbar_IsDerelict);
		eval("t.$Towbar_IsDerelict = function(entity) { \
			var w = worldScripts.escortdeck; \
			if( (!w.$EscortDeckXL || \
				w.$EscortDeckCWS && player.ship && player.ship.isValid \
				&& player.ship.dataKey.indexOf('carrier') > -1 ) && \
				worldScripts.towbar.$EscortDeckToWS_IsDerelictOrig(entity) \
				&& ( !w.$EscortDeckShip \
				|| w.$EscortDeckShip.indexOf(entity) == -1 ) ) \
				 return true; \
			return false; \
		}");
		eval("t.$EscortDeckToWS_shipCollidedOrig = "+t.shipCollided);
		eval("t.shipCollided = function(otherShip) { \
			var w = worldScripts.escortdeck; \
			if( (!w.$EscortDeckXL || \
				w.$EscortDeckCWS && player.ship && player.ship.isValid && \
				player.ship.dataKey.indexOf('carrier') > -1 ) && \
				!w.$EscortDeckShip || w.$EscortDeckShip.indexOf(otherShip) == -1 ) \
				worldScripts.towbar.$EscortDeckToWS_shipCollidedOrig(otherShip); \
		}");
	}
}

this.startUpComplete = function() {
	//call after towbar and ccl startups finished
	this.$EscortDeck_SalvageBadEscorts(player.ship, this);
	this.$EscortDeck_Interface();
}

this.alertConditionChanged = function(newCondition, oldCondition) {
	if( !player.ship || !player.ship.isValid ) return;
 	if( newCondition == 3 && player.alertHostiles ) {
		this.$EscortDeck_ControlAll(true, false, this); //launch all escorts
	} else if( newCondition == 1 ) { //call back escorts in green alert
		this.$EscortDeck_ControlAll(false, false, this);
	}
}

this.missionScreenOpportunity = function() {
	var t = this.$EscortDeckToWS; //use Towbar salvage if available
	if( player.ship.docked && this.$EscortDeckFirstMSO
		&& this.$EscortDeckShip && this.$EscortDeckShip.length > 0 ) {
		this.$EscortDeckFirstMSO = false; //do only once
		if( t && t.$TowbarShip && t.$TowbarShip.isValid
			&& !t.$Towbar_IsHeavyShip( t.$TowbarShip ) ) { //avoid heavy ships in aegis with autodock
			t.$Towbar_Payout(t.$TowbarShip); //call payout in towbar oxp
		}
		for(var i = 0; i < this.$EscortDeckPadPos.length; i++ ) { //salvage unusables
			var s = this.$EscortDeckShip[i]; //the ship on this pad
			if( s && s.isValid ) {
				if( !s.script || !( s.script.$EscortDeckUsable > 0 ) ) { //unusable
					this.$EscortDeck_Salvage(i, this, player.ship); //salvage
					if(this.$EscortDeckToWS) {
						this.$EscortDeckFirstMSO = true; //do one at a time
						i = this.$EscortDeckPadPos.length;
					}
				}
			} else { //inside stations check saved ship data
				var d = this.$EscortDeckShipData[i];
				if( d && d[1] && ( !d[14] || !(d[14].usable > 0) ) ) {//unusable
					this.$EscortDeck_Salvage(i, this, player.ship);
					if(this.$EscortDeckToWS) {
						this.$EscortDeckFirstMSO = true; //do one at a time
						i = this.$EscortDeckPadPos.length;
					}
				}
			}
		}
	}
	if( t && t.$TowbarRunScreens.length > 0 ) {
		var rs = t.$TowbarRunScreens.shift(); //get and remove the oldest element
		if(this.$Debug) log(this.name, "MSO "+t.$TowbarRunScreens.length+" "+guiScreen+" "+rs );
		t.$Towbar_RunScreen(rs.place, rs.msg, rs.model); //show sell salvage screen
	}
}

this.playerBoughtEquipment = function(equipment) {
	var p = player.ship;
	var w = this; //w is used below
	if( equipment.indexOf("EQ_ESCORTDECK_" ) == 0 ) {
		p.removeEquipment(equipment);
		var k = equipment.substring(14); //extract datakey from the end of the eq key
		var s = this.$EscortDeck_SpawnOne( "[escortdeck-"+k+"]", this );
		if(this.$Debug) log(this.name, k+" "+s); //debug
		if( !s || !s[0] || !s[0].isValid ) { //possible bug
			this.$EscortDeck_RollBack(equipment, "You can't buy "+k+".");
			return;
		} else s = s[0];
		var c = EquipmentInfo.infoForKey(equipment).price/10; //escort price
		var en = EquipmentInfo.infoForKey(equipment).name; //escort name
		w.$EscortDeck_AwardEQs(s);
		var d = this.$EscortDeck_MakeShipData(s);
		if(d[14]) d[14].usable = 1; //usable
		var added = false;
		var reason = this.$EscortDeck_TooLarge(s, p, this);
		if( !reason && !this.$EscortDeckXL && this.$EscortDeck_isLarge(p, this)
			&& this.$EscortDeck_isMini(s) ) { //Adder or smaller, try side decks first
			for( var i = 0; !added && i < w.$EscortDeckPadPos.length; i++ ) {
				if( w.$EscortDeck_isMiniPad(i, p, w) ) {
					if( !this.$EscortDeckShipData[i] ) {
						this.$EscortDeckShipData[i] = d;
						added = i+1; //=pad+1
					}
				}
			}
		}
		var len = this.$EscortDeckPadPos.length;
		if( !reason && !added ) for(var i = 0; i < len; i++ ) {
			if( !this.$EscortDeckShipData[i] 
				&& !w.$EscortDeck_TooLarge2(s, p, w, i) ) {
				this.$EscortDeckShipData[i] = d;
				added = i+1;
				i=10;
			}
		}
		s.remove(true);
		if( added ) player.consoleMessage("You paid "+c
			+" credits for "+en+" placed to landing pad "+added, 10);
		else {
			if(!reason) reason = "no room on your deck";
			this.$EscortDeck_RollBack( equipment, "You can't buy "+en
			+", "+reason );
		}
	}
	if( equipment == "EQ_ESCORTDECK" ) {
		this.$EscortDeckXL = false;
		if( w && w.$EscortDeckCWS && p.dataKey.indexOf("carrier") > -1
			&& p.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" ) {
			//salvage pad 8, exchange pad 2-3 and shift pad 9-10 to 8-9
			w.$EscortDeck_Salvage(7, w, p); //salvage a ship to reduce from 10 to 9
			var d = [], s = [];
			d[0] = w.$EscortDeckShipData[0]; s[0] = w.$EscortDeckShip[0];
			d[1] = w.$EscortDeckShipData[2]; s[1] = w.$EscortDeckShip[2];
			d[2] = w.$EscortDeckShipData[1]; s[2] = w.$EscortDeckShip[1];
			d[3] = w.$EscortDeckShipData[3]; s[3] = w.$EscortDeckShip[3];
			d[4] = w.$EscortDeckShipData[4]; s[4] = w.$EscortDeckShip[4];
			d[5] = w.$EscortDeckShipData[5]; s[5] = w.$EscortDeckShip[5];
			d[6] = w.$EscortDeckShipData[6]; s[6] = w.$EscortDeckShip[6];
			d[7] = w.$EscortDeckShipData[8]; s[7] = w.$EscortDeckShip[8];
			d[8] = w.$EscortDeckShipData[9]; s[8] = w.$EscortDeckShip[9];
			w.$EscortDeckShipData = d;
			w.$EscortDeckShip = s;
		}
		p.removeEquipment("EQ_ESCORTDECKXL"); //replace
		this.$EscortDeck_SalvageBadEscorts(p, this);
	}
	if( equipment == "EQ_ESCORTDECKXL" ) {
		this.$EscortDeckXL = true;
		if( w && w.$EscortDeckCWS && p.dataKey.indexOf("carrier") > -1
			&& p.equipmentStatus("EQ_ESCORTDECK") == "EQUIPMENT_OK" ) {
			//exchange pad 2-3 and shift pad 8-9 to 9-10
			var d = [], s = [];
			d[0] = w.$EscortDeckShipData[0]; s[0] = w.$EscortDeckShip[0];
			d[1] = w.$EscortDeckShipData[2]; s[1] = w.$EscortDeckShip[2];
			d[2] = w.$EscortDeckShipData[1]; s[2] = w.$EscortDeckShip[1];
			d[3] = w.$EscortDeckShipData[3]; s[3] = w.$EscortDeckShip[3];
			d[4] = w.$EscortDeckShipData[4]; s[4] = w.$EscortDeckShip[4];
			d[5] = w.$EscortDeckShipData[5]; s[5] = w.$EscortDeckShip[5];
			d[6] = w.$EscortDeckShipData[6]; s[6] = w.$EscortDeckShip[6];
			d[7] = null; s[7] = null;
			d[8] = w.$EscortDeckShipData[7]; s[8] = w.$EscortDeckShip[7];
			d[9] = w.$EscortDeckShipData[8]; s[9] = w.$EscortDeckShip[8];
			w.$EscortDeckShipData = d;
			w.$EscortDeckShip = s;
		}
		p.removeEquipment("EQ_ESCORTDECK"); //replace
		this.$EscortDeck_SalvageBadEscorts(p, this);
	}
	if( equipment == "EQ_FUEL" ) {
		var d = this.$EscortDeckShipData;
		var f = 0;
		var s = "";
		var len = this.$EscortDeckPadPos.length;
		if(d) for(var i = 0; i < len; i++ ) {
			if( d[i] && d[i][9] < 7 ) {
				if( f > 0 ) s = "s"; //more than one escort ship need refuel
				f += 7 - d[i][9];
			}
		}
		if( f > 0 ) {
			var c = f * 3;//3cr/ly
			if( player.credits > c ) {
				for(var i = 0; i < len; i++ ) {
					if( d[i] && d[i][9] < 7 ) d[i][9] = 7;
				}
				player.credits -= c;
				player.consoleMessage("Your escort"+s+" got "+(Math.round(f*10)/10)
					+" ly fuel for "+(Math.round(c*10)/10)+" credits", 10);
			}
		}
	}
}

this.playerBoughtNewShip = function(ship) {
	this.$EscortDeck_SaveMaxs( player.ship, this );
	if( !this.$EscortDeckCWS || !this.$EscortDeckCWS.$CarriersPlayerInEscort ) {
		this.$EscortDeck_SalvageBadEscorts(player.ship, this);
		//setup new pads after SalvageBadEscorts only!
		this.$EscortDeck_SetPads( player.ship, this );
	}
}

this.playerStartedJumpCountdown = function(type, seconds) {
	//fix for Granite ships to the too close high mass do not prevent the jump
	if( !this.$EscortDeckCWS || player.ship.dataKey.indexOf("carrier") == -1 ) {
		for( var i = 0; i < this.$EscortDeckPadPos.length; i++ )
			if( this.$EscortDeckShipPos[i] )
				this.$EscortDeckShipPos[i].z -= 40; //move backward on non-Carriers
	} else { //on Carriers move escort up
		for( var i = 0; i < this.$EscortDeckPadPos.length; i++ )
			if( this.$EscortDeckShipPos[i] )
				this.$EscortDeckShipPos[i].y += 100;
	}
}

this.playerWillSaveGame = function(message) {
	missionVariables.$EscortDeckShipData = JSON.stringify(this.$EscortDeckShipData);
	missionVariables.$EscortDeckLocked = JSON.stringify(this.$EscortDeckLocked);
}

this.shipAttackedOther = function(other) { //player hits other
	if( !other.script ) other.script = "oolite-default-ship-script.js"; //for sure
	if( other.script ) other.script.$EscortDeck_Target = 1; //flag for escort AI

	if( this.$EscortDeckToWS //a copy of derelict maker code from towbar
		|| !other || !other.isValid || other.isPlayer ) return; //for sure
	if( !other.isDerelict && other.energy < 23 && other.energy > 0 && !other.isThargoid ) {
		//the next Military Laser hit can destroy the target
		other.fireMissile();//last resort
		if( other.equipmentStatus("EQ_ESCAPE_POD") === "EQUIPMENT_OK"
			&& other.abandonShip()) {//eject to make more derelict ship but only if has Escape Pod
			if( worldScripts["detectors"] ) {
				var dn = other.script.$Detectors_Origname;//short name with version
			} else {
				var dn = other.displayName;
				other.displayName = "Derelict " + other.displayName;
			}
			player.consoleMessage(dn+" ejected!");
			if( other.script && !other.script.$AlreadyScored ) {
				other.script.$AlreadyScored = true;
				player.score++;
			}
		}
	}
	if( other.isDerelict && !other.isThargoid ) {
		this.$EscortDeckSound3.play(); //audio feedback if derelict
		if( other.script ) {
			if( !other.script.$TowbarDerelictAttack ) {
				other.script.$TowbarDerelictAttack = 1;
			} else ++other.script.$TowbarDerelictAttack;
		}
		if( !other.script || other.script.$TowbarDerelictAttack <= 10 ) { //prevent invulnerability
			other.energy += 14;//avoid destroying the target when just derelicted
		}
	}
	this.$EscortDeck_Bounty( other );
}

this.shipCollided = function(otherShip) {
	var p = player.ship;
	if( !p || !p.isValid ) return;
	var w = this;
	var i = w.$EscortDeckShip.indexOf(otherShip);
	if( i > -1 ) {
		if( w.$EscortDeckShipPos[i] ) { //move escort until avoid collision
			if( !w.$EscortDeckCWS || p.dataKey.indexOf("carrier") == -1 ) {
				w.$EscortDeckShipPos[i].z -= 2; //move backward on non-Carriers
			} else { //on Carriers move escort up until avoid collision
				w.$EscortDeckShipPos[i].y += 5;
//				if( w.$EscortDeckXL ) {
//					if( i < 4 ) w.$EscortDeckShipPos[i].y += 5; //top pads up
//					else w.$EscortDeckShipPos[i].y -= 5; //others move down
//					else if( Math.round(i/2) != i/2 )
//						w.$EscortDeckShipPos[i].x += 5; //move right
//					else w.$EscortDeckShipPos[i].x -= 5; //move left
//				} else {
//					if( i < 5 ) w.$EscortDeckShipPos[i].y += 5; //on top move up
//					else w.$EscortDeckShipPos[i].y -= 5; //others move down
//				}
			}
		}
		if( otherShip.AIState == "LANDING" ) {
			var r = w.$EscortDeck_PutShip(otherShip, i, p, w, true, true);
			if( !r ) otherShip.AIState == "DECK"; //landed
			else {
				otherShip.AIState == "FOLLOW"; //no room?
				if(w.$Debug) log(w.name, "Can't add "+t.displayName
				+" to Escort Deck pad "+i+", "+r); //debug
			}
		}
	} else w.$EscortDeck_AddEscort( otherShip, w.$EscortDeckShip, p, w, true );
}

this.shipDockedWithStation = function(station) {
	this.$EscortDeckDockingTunnel = false;
	this.$EscortDeck_Interface();
}

this.shipExitedWitchspace = function() {
	this.$EscortDeck_Show(player.ship);
	this.$EscortDeckDockingTunnel = false;
}

this.shipLaunchedFromStation = function() {
	this.$EscortDeckDockingTunnel = false;
	if( this.$EscortDeckCWS && this.$EscortDeckCWS.$CarriersPlayerInEscort )
		return;
	this.$EscortDeck_Show(player.ship);
}

this.shipTargetAcquired = function(t) {  //make some derelicts usable
	if( !t || !t.isValid ) return;
	var p = player.ship;
	var pad = this.$EscortDeckShip.indexOf(t);
//	if( pad > -1 ) this.$EscortDeckSelectedPad = pad; - do not use, cause disturbing changes in selected pad
	if( !t.script ) t.script = "oolite-default-ship-script.js"; //for sure
	if( t.script && pad == -1  && ( t.isDerelict || t.script.$EscortDeckUsable > 0 )
		&& (!this.$EscortDeck_isLarge(t, this) //exclude ships over 130t except for Carriers
			|| this.$EscortDeckCWS && p.dataKey.indexOf("carrier") > -1
			&& t.mass < this.$EscortDeckCMaxMass ) ) { //must below this mass on Carriers
		if( t.script.$TowbarMinedShip ) t.script.$EscortDeckUsable = -1; //unusable as escort
		var reason = this.$EscortDeck_TooLarge(t, p, this); //fit into the deck limits?
		var rnd = 0.1; //chance to find an undamaged derelict ship
		if( !(t.script.$EscortDeckUsable > 0) && !(t.script.$EscortDeckUsable < 0) ) {
			//determine usability once only
			if( !reason && Math.random() < rnd ) { //fit into the deck limits and lucky
				t.script.$EscortDeckUsable = 1;
				this.$EscortDeck_AwardEQs(t);
			} else {
				t.script.$EscortDeckUsable = -1; //unusable as escort
				t.script.$TowbarUsableShip = false;
				if(this.$Debug) log(this.name, t.displayName+" is set to unusable: "+t);
			}
		}
		var shipmass = t.mass/1000;
		//remove extra density from HardShips and Granite ships
		if( t.scriptInfo && t.scriptInfo.hardarmour > 1 )
			shipmass = shipmass / 2;
		if( t.script.$EscortDeckUsable > 0 ) {
			t.script.$TowbarUsableShip = true; //show Usable! flag in CombatMFD
			var f = "Found an usable "+t.name+" ("+Math.ceil(shipmass)+"t) ";
			var g = null;
			if( p.equipmentStatus("EQ_ESCORTDECK") != "EQUIPMENT_OK"
				 && p.equipmentStatus("EQ_ESCORTDECKXL") != "EQUIPMENT_OK" )
				g = "You need an Escort Deck equipment to use as your escort";
			else if( reason ) f = t.name+" is "+reason+" for your Escort Deck";
			else if( t.position.distanceTo(p.position) > this.$EscortDeckLandingDist
				 + t.collisionRadius )
				g = "Go near to pick up to the Escort Deck";
			if(this.$Debug) log(this.name, f+" "+g); //debug
			player.consoleMessage(f, 10);
			if(g) player.consoleMessage(g, 10);
		} else if( t.script.$EscortDeckUsable < 0 ) {
			t.script.$TowbarUsableShip = false;
			if( reason ) {
				player.consoleMessage(t.name+" is "+reason, 4.5);
				t.script.$EscortDeckUsable = -2; //flag to addescort
			} else player.consoleMessage("Salvage "+t.name
				+" ("+Math.ceil(shipmass)+"t)", 4.5);
		}
	}
	if( pad == -1 && t.script && t.script.$EscortDeckUsable > 0 )
		p.awardEquipment("EQ_DTAUSA"); //show Usable label on NumericHUD
	else p.removeEquipment("EQ_DTAUSA"); //remove Usable label on NumericHUD
}

this.shipTargetLost = function() {
	player.ship.removeEquipment("EQ_DTAUSA"); //remove Usable label on NumericHUD
}

this.shipWillDockWithStation = function(station) {
	this.$EscortDeckDockingTunnel = true;
	if( station.dataKey != "carriers-vdock" 
		&& station.dataKey.indexOf("carrier-player") == -1 )
		this.$EscortDeck_VClear(this);
	else this.$EscortDeckFirstMSO = false; //do not salvage in vdock
	player.ship.removeEquipment("EQ_DTAUSA"); //remove Usable label on NumericHUD
}

this.shipWillEnterWitchspace = function(cause) {
	this.$EscortDeckDockingTunnel = true;
	this.$EscortDeck_VClear(this);
	player.ship.removeEquipment("EQ_DTAUSA"); //remove Usable label on NumericHUD
}

this.shipWillLaunchFromStation = function() {
	this.$EscortDeckDockingTunnel = true;
//	this.$EscortDeck_Show(player.ship); - lose escorts here, must call in shipLaunchedFromStation
	var ps = player.ship; //show mfd as early as other mfds but avoid no escorts message
	if( ps.setMultiFunctionText ) ps.setMultiFunctionText(this.name, "", false);
	
	if( this.$EscortDeckCWS && this.$EscortDeckCWS.$CarriersPlayerInEscort )
		return;
	
	this.$EscortDeckFirstMSO = true;
	if( this.$Derelicts && system && !system.isInterstellarSpace ) { //derelict targets
		var m = system.mainStation;
		var r = 300; //spread ship in this radius
		var ships = system.addShips("escort", 16, //random escorts over the nav buoy
			m.position.add(m.vectorForward.multiply( 11000 )), 3*r);

//		var a = system.addShips("[anaconda]", 2,
//			m.position.add(m.vectorForward.multiply( 2500 )), r);
//		ships.push(a[0]); ships.push(a[1]);
		var a = system.addShips("[escortdeck-adder]", 4,
			m.position.add(m.vectorForward.multiply( 3000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
/*		var a = system.addShips("[escortdeck-asp]", 4,
			m.position.add(m.vectorForward.multiply( 3500 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[escortdeck-gecko]", 4,
			m.position.add(m.vectorForward.multiply( 4000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[escortdeck-sidewinder]", 4,
			m.position.add(m.vectorForward.multiply( 4500 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[escortdeck-viper]", 4,
			m.position.add(m.vectorForward.multiply( 5000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		
		var a = system.addShips("[cobramk1]", 4,
			m.position.add(m.vectorForward.multiply( 6000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[ferdelance]", 4,
			m.position.add(m.vectorForward.multiply( 7000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[moray]", 4,
			m.position.add(m.vectorForward.multiply( 8000 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[cobra3-player]", 4,
			m.position.add(m.vectorForward.multiply( 6500 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
		var a = system.addShips("[krait]", 4,
			m.position.add(m.vectorForward.multiply( 7500 )), r);
		ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
*/

		if( this.$EscortDeckPWS ) {
			var a = system.addShips("[escortdeck-ophidian]", 4,
				m.position.add(m.vectorForward.multiply( 8000 )), r);
			ships.push(a[0]); ships.push(a[1]); ships.push(a[2]); ships.push(a[3]);
/*			var a = system.addShips("[escortdeck-ghavial]", 2,
				m.position.add(m.vectorForward.multiply( 8000 )), r);
			ships.push(a[0]); ships.push(a[1]);
			var a = system.addShips("[escortdeck-wolf]", 2,
				m.position.add(m.vectorForward.multiply( 8500 )), r);
			ships.push(a[0]); ships.push(a[1]);
			var a = system.addShips("[escortdeck-sniper-wolf]", 2,
				m.position.add(m.vectorForward.multiply( 8500 )), r);
			ships.push(a[0]); ships.push(a[1]);
*/			var a = system.addShips("[escortdeck-arachnid]", 2,
				m.position.add(m.vectorForward.multiply( 9000 )), r);
			ships.push(a[0]); ships.push(a[1]);
			var a = system.addShips("[escortdeck-sniper-arachnid]", 2,
				m.position.add(m.vectorForward.multiply( 9000 )), r);
			ships.push(a[0]); ships.push(a[1]);
			var a = system.addShips("[escortdeck-s8]", 2,
				m.position.add(m.vectorForward.multiply( 9500 )), r);
			ships.push(a[0]); ships.push(a[1]);
			var a = system.addShips("[escortdeck-sniper-s8]", 2,
				m.position.add(m.vectorForward.multiply( 9500 )), r);
			ships.push(a[0]); ships.push(a[1]);
		}

		for(var i=0; i<ships.length; i++) {
			if( ships[i] ) {
				ships[i].orientation = ps.orientation;
				ships[i].switchAI("nullAI.plist"); //need to deactivate escortdeck AI
				if( worldScripts["shipversion"] ) {
					worldScripts["shipversion"].shipSpawned(ships[i]);
				}
				ships[i].awardEquipment("EQ_ESCAPE_POD");
				if(i == 0) { //make first ship to weak offender without injectors, need a few shot only to eject it
					ships[i].removeEquipment("EQ_FUEL_INJECTION");
					ships[i].bounty=33;
					ships[i].energy=33;
				} else ships[i].abandonShip(); //other ships derelict
			} else if( this.$Debug ) log(this.name, "ships["+i+"] is undefined ");
		}
		if( this.$Debug ) log(this.name, "test derelict ships: "+ships);
	}
}

//EscortDeck methods
this.$EscortDeck_AddEscort = function( t, e, owner, w, msg ) { //add t into the deck of ship if possible
	if( t && t.isValid && t.script &&
		( t.isDerelict || t.script.$EscortDeckUsable == -1 || t.script.$EscortDeckUsable > 0 )
		&& t.mass > 1000 && t.mass < this.$EscortDeckExtremeMass && !t.isVisualEffect && !t.isPlayer
		&& !t.isDock && !t.isStation && !t.isCargo && !t.isRock && !t.isThargoid && !t.isWeapon
		&& owner && owner.isValid
		&& ( owner.equipmentStatus("EQ_ESCORTDECK") == "EQUIPMENT_OK"
			|| owner.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" )
		&& owner.mass >= w.$EscortDeckMinMass
		&& owner.position.distanceTo(t.position) < w.$EscortDeckLandingDist + t.collisionRadius ) {
		var p = e.indexOf(t);
		if( p > -1 && t.AIState != "LANDING" ) return false;//owned escort is on way, do not land
	
		var r = this.$EscortDeck_TooLarge(t, owner, w);
		if( r ) { //reason
			if( msg && owner == player.ship ) {
				player.consoleMessage("You can't pick up "+t.displayName
					+" to your Escort Deck, "+r, 10);
			}
			return false;
		}

		var diff = owner.velocity.subtract( t.velocity ).magnitude();
		if( diff > 100 ) {
			if( owner == player.ship && msg && p == -1 ) //not in auto landing
				player.consoleMessage(Math.round(diff/10)*10+" m/s speed difference");
			return false;
		}

		//put into the selected pad if fit into
		//and regardless of occupied (replace) if this is the current target
		var spad = w.$EscortDeckSelectedPad;
		if( spad > -1 ) {
			if(this.$Debug) log(w.name, "Try "+t.displayName
				+" to selected pad "+(spad+1)); //debug
			var f = false; //force replace
			if( owner.target == t ) f = true;
			r = w.$EscortDeck_PutShip(t, spad, owner, w, msg, f);
			if( !r ) return true; //report successfully added
			else if(this.$Debug) log(w.name, "Can't add "+t.displayName
				+" to selected pad "+(spad+1)+", "+r); //debug
		}

		//put owned escorts back to the correct pad
		if( p > -1 ) {
			if(this.$Debug) log(w.name, "Try "+t.displayName
				+" to old pad "+(p+1)); //debug
			r = w.$EscortDeck_PutShip(t, p, owner, w, msg, true);
			if( !r ) return true; //report successfully added
			else { //if faliled the remove and try other pads as new ships
				if(this.$Debug) log(w.name, "Can't add "+t.displayName
					+" back to pad "+(p+1)+", "+r); //debug
				e[p] = null;
				p = -1;
			}
		}
		if(this.$Debug) log(w.name, "Try "+t.displayName+" to empty pads"); //debug

		//if Adder or smaller then try Adder pads first
		var large = w.$EscortDeck_isLarge(owner, w);
		if( large && //side pads need large ship
			w.$EscortDeck_AddMini(t, owner, w, msg, false) )
			return true; //report successfully added

		//non-small escort need large ship
		for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
			if( !w.$EscortDeck_isMiniPad(i, owner, w) ) {
				//find the first empty pad except Adder only pads
//				log(w.name, i+". free:"+w.$EscortDeck_isPadFree(i, w)); //debug
				r = w.$EscortDeck_PutShip(t, i, owner, w, msg, false);
				if( !r ) i = w.$EscortDeckPadPos.length;//exit from the for cycle
			}
		}
		if( e.indexOf(t) > -1 ) return true; //report successfully added

		//replace first flying escort if no more empty pad and none selected
		if(this.$Debug) log(w.name, "Try "+t.displayName+" to replace a flying escort");
		for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
			if( !w.$EscortDeck_isMiniPad(i, owner, w)
				|| large && w.$EscortDeck_isMini(t) ) {
				if( !w.$EscortDeckShipPos[i] ) { //on the way
					r = w.$EscortDeck_PutShip(t, i, owner, w, msg, true);
					if( !r ) return true;
				}
			}
		}

		if( e.indexOf(t) > -1 ) return true; //report successfully added
		else {
			if( msg && owner == player.ship ) {
				player.consoleMessage("No room on Escort Deck for "
					+t.displayName, 10);
			}
			if(this.$Debug) log(w.name, "Can't add "+t.displayName
				+" to Escort Deck, "+r); //debug
			return false;
		}
	}
	return false; //can not add escort to the deck
}

this.$EscortDeck_AddMini = function( t, owner, w, msg, land ) { //add mini escort to a small pad if possible
	if( !w.$EscortDeck_isMini(t) ) return false; //need like an Adder or smaller
	for( var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
		if( w.$EscortDeck_isMiniPad(i, owner, w) ) {
			if( !w.$EscortDeck_PutShip(t, i, owner, w, msg, land ) )
				return true;
		}
	}
	return false;
}

this.$EscortDeck_AwardEQs = function( s ) {
	//award player specific equipments, useful when sitting in from a Carrier
	if( s && s.isValid ) {
		s.awardEquipment("EQ_ADVANCED_COMPASS");
		if(worldScripts.combat_MFD) s.awardEquipment("EQ_COMBATMFD");
		s.awardEquipment("EQ_ESCAPE_POD");
		s.awardEquipment("EQ_MULTI_TARGET");
		s.awardEquipment("EQ_SCANNER_SHOW_MISSILE_TARGET");
		if(worldScripts.targetAutolock) //requires EQ_SCANNER_SHOW_MISSILE_TARGET so add after
			s.awardEquipment("EQ_TARGET_AUTOLOCK");
		s.awardEquipment("EQ_TARGET_MEMORY");
		s.awardEquipment("EQ_INTEGRATED_TARGETING_SYSTEM");//requires Target Memory so add after
		if(worldScripts.telescope) s.awardEquipment("EQ_TELESCOPE");
		s.awardEquipment("EQ_WORMHOLE_SCANNER");
		s.awardEquipment("EQ_HEAT_SHIELD");//protection for sunskimming, requested by QCS
	}
}

this.$EscortDeck_Bounty = function( ship ) {
	if( ship.isDerelict && ship.bounty > 0 ) {
		player.credits += ship.bounty;
		player.consoleMessage("Bounty: "+formatCredits(ship.bounty, true, true));
		player.consoleMessage("Total: "+formatCredits(player.credits, true, true));
		ship.bounty = 0;
	}
}

this.$EscortDeck_ControlAll = function( launch, manual, w ) { //launch or call back all escorts
	if( launch && w.$EscortDeck_ControlFast(manual) ) return(false);
	var c = 0;
	var carrier = false;
	if( w.$EscortDeckCWS && player.ship.dataKey.indexOf("carrier") > -1 )
		carrier = true;
	for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
		if( !w.$EscortDeckLocked[i] &&
			( !launch || !carrier || w.$EscortDeckXL || manual ||
			//can prevent auto launch of Anaconda from Carrier
			w.$EscortDeckCBigTraderLaunch ||
			w.$EscortDeckShip && w.$EscortDeckShip[i]
			&& w.$EscortDeckShip[i].mass < w.$EscortDeckXLMass ) ) {
			if( w.$EscortDeck_Control1( launch, i, false, w, manual ) ) //launch without message
				c++; //count usable escorts
		}
	}
	if( c > 0 && ( manual || !launch ) ) { //do not message autolaunch when entering red alert
		var e = c+" Escorts ";
		if( c == 1 ) e = "1 Escort ";
		var m = "called back";
		if( launch ) m = "on the way";
		player.consoleMessage(e+m, 10);
	}
	w.$EscortDeck_MFD(w); //update MFD
	return(c);
}

this.$EscortDeck_ControlFast = function(manual) {
	var max = player.ship.maxSpeed; //no launch over max.speed
//	if( !manual ) max = max * 0.999; //no autolaunch at max.speed
	if( player.ship.speed > max ) {
		if( manual ) player.consoleMessage("You travel too fast to launch escorts", 3);
//		else player.consoleMessage("No autolaunch of escorts at maximal speed", 3);
		return(true);
	}
	return(false);
}

this.$EscortDeck_Control = function( launch, pad, msg, w ) { //launch or call back an escort
	w.$EscortDeck_Control1( launch, pad, msg, w, true );
}

this.$EscortDeck_Control1 = function( launch, pad, msg, w, manual ) { //launch or call back an escort
	if( launch && this.$EscortDeck_ControlFast(manual) ) return false;
	var s = null;
	var p = player.ship;
	if( w.$EscortDeckShip ) s = w.$EscortDeckShip[pad];
	if( s && s.isValid ) {
		if( !s.script || !( s.script.$EscortDeckUsable > 0 ) ) { //unusable
			if( msg ) w.$EscortDeck_Release(p, pad, w); //msg mean manual and one ship
			return(false);
		}
		var m = null; //the carrier can't launch
		if( launch && !w.$EscortDeckLocked[pad] ) {
			if( p.removeCollisionException ) //from v1.81
				p.removeCollisionException(s);
			w.$EscortDeckShipData[pad] = null; //prevent recreation in dock
			w.$EscortDeckShipPos[pad] = null; //do not update position in FCB

			//increase steering and acceleration in proportion with the smaller mass on deck
			w.$EscortDeck_SetMaxs( p, w );

			if( s.AI != "EscortDeck_AI.plist" ) s.switchAI("EscortDeck_AI.plist");
			s.AIState = "GOTO"; //switch back from waypoint AI
			s.scannerDisplayColor1 = [0, 0.7, 0]; //green
			s.velocity = p.velocity; //prevent collision if player travel at speed
			if(w.$EscortDeckPadVec) {//help to leave the owner: jump 20m far
				var j = w.$EscortDeckJumpDist;
				var x = j * w.$EscortDeckPadVec[pad][0];
				var y = j * w.$EscortDeckPadVec[pad][1];
				var z = j * w.$EscortDeckPadVec[pad][2];
				if( p.dataKey.indexOf("carrier") > -1 ) {
					if( w.$EscortDeck_isMiniPad(pad, p, w) ) {
						z = j; //Carrier mini pads jump forward
					} else if( !w.$EscortDeckXL ) { //traders jump sideways
						if( pad == 6 ) {x = j; z = 0;}
						else if( pad == 5 ) {x = -j; z = 0;}
					}
				}
				var o = p.orientation;
				var sp = s.position.add(o.vectorRight().multiply(x))
					.add(o.vectorUp().multiply(y))
					.add(p.heading.multiply(z));
				if( sp.magnitude() > 500 ) s.position = sp;
			}
			m = "on the way";
		} else { //call back
			if( s.AIState != "DECK" ) { //isn't in deck already
//			if( !w.$EscortDeckShipPos[pad] ) {
				if( s.AI != "EscortDeck_AI.plist" ) s.switchAI("EscortDeck_AI.plist");
				s.AIState = "LANDING";  //switch back from waypoint AI
				m = "called back";
			}
		}
		if( m && msg ) {
			player.consoleMessage(s.displayName+" "+m, 5);
			w.$EscortDeck_MFD(w); //update MFD
		}
		if(this.$Debug) log(w.name, pad+". launch:"+launch+" "+m); //debug
		if( m ) return true;
	}
	return false;
}

this.$EscortDeck_EQActivated = function(w) { //launch or call back one or all escorts
	var p = player.ship;			//called from carriers-escort-eq.js also!
	if( !p || !p.isValid ) return;
	if( w.$EscortDeckSelectedPad == w.$EscortDeckPadPos.length && w.$EscortDeckToWS ) {
		if( w.$EscortDeckToWS.$TowbarShip ) w.$EscortDeckToWS.$Towbar_Release();
		else player.consoleMessage("No ship on Towbar");
	} else if( w.$EscortDeckSelectedPad == -2 ) { //setup sit into mode
		w.$EscortDeckSitInto = !w.$EscortDeckSitInto; //flip-flop
	 	w.$EscortDeck_EQDisplayMode(w);
	} else if( w.$EscortDeckSelectedPad == -1 ) { //launch or call back all escorts
		w.$EscortDeckLaunch = !w.$EscortDeckLaunch; //flip-flop
		w.$EscortDeck_ControlAll(w.$EscortDeckLaunch, true, w);
	} else { //launch or call back a selected escort with message
		var s = w.$EscortDeckShip[w.$EscortDeckSelectedPad];
		if( s && s.isValid ) {
			if( w.$EscortDeckShipPos[w.$EscortDeckSelectedPad] ) {
				if( w.$EscortDeckLocked[w.$EscortDeckSelectedPad] ) {
					w.$EscortDeckLaunch = true; //launch if on deck and locked
					w.$EscortDeckLocked[w.$EscortDeckSelectedPad] = false;
				} else {
					w.$EscortDeckLaunch = false;
					w.$EscortDeckLocked[w.$EscortDeckSelectedPad] = true; //lock
				}
				w.$EscortDeck_MFD(w); //update MFD
			} else {
				w.$EscortDeckLocked[w.$EscortDeckSelectedPad] = false;
				if( s.AIState == "LANDING" ) w.$EscortDeckLaunch = true;
				else {
					w.$EscortDeckLaunch = false;
					if( s.dataKey.indexOf("carrier-player") > -1 ) {
						if( p.target == s ) {//target is the player's carrier?
							player.consoleMessage("Approach the Carrier "+
								"within "+w.$EscortDeckLandingDist+
								"m with small speed difference "+
								" for landing",10);
						} else { player.consoleMessage("Hold ident lock on"+
								" your Carrier to keep up"+
								" landing clearance",5);
						}
					}
				}
			}
			if( w.$EscortDeckSitInto //deck is in the "sit into the escort" mode
				&& w.$EscortDeckCWS //Carriers OXP is installed
				&& p.dataKey.indexOf("carrier-player") > -1 //player in the Carrier
				&& s.AIState == "DECK" //selected a landed, usable and playable escort
				&& s.missileCapacity > 0 //bugfix against a core bug with 0 max_missiles
				&& w.$EscortDeckPlayableDataKeys.indexOf(s.dataKey) > -1 ) {
				w.$EscortDeckCWS.$Carriers_SitIntoAnEscort(s); //launch player
			} else w.$EscortDeck_Control( w.$EscortDeckLaunch,
				w.$EscortDeckSelectedPad, true, w);
		} else { //no ship in this pad
			if( w.$EscortDeckLocked[w.$EscortDeckSelectedPad] )
				w.$EscortDeckLocked[w.$EscortDeckSelectedPad] = false;
			else w.$EscortDeckLocked[w.$EscortDeckSelectedPad] = true;
		}
	}
}

this.$EscortDeck_EQDisplayMode = function(w) {
 	if( w.$EscortDeckSelectedPad == -1 ) { //launch or callback mode
		var l = "Launch all non-locked";
		if( w.$EscortDeckLaunch ) l = "Call back all flying";
		player.consoleMessage(l+" escorts when EscortDeck activated",5);
	} else if( w.$EscortDeckSelectedPad == -2 ) { //setup sit into mode
		var l = "avoid";
		if( w.$EscortDeckSitInto ) l = "allow";
		player.consoleMessage("EscortDeck "+l+" to sit into a selected escort",5);
	}
}

this.$EscortDeck_EQMode = function(w, playable) { //select the next pad on escort deck
 	var p = player.ship;		 //called from carriers-escort-eq.js also!
	if( p && p.isValid ) {
		w.$EscortDeckSelectedPad++;
/*		if( playable ) { - removed to allow select any ship to drop salvage, etc.
			var start = false;
			if( w.$EscortDeckSelectedPad == 0 ) start = true; //start from the first pad
			var found = -1;
			while( found < 0 && w.$EscortDeckSelectedPad < w.$EscortDeckPadPos.length ) {
				var s = w.$EscortDeckShip[w.$EscortDeckSelectedPad];
				if( s && s.isValid && s.script && s.script.$EscortDeckUsable > 0
					&& w.$EscortDeckPlayableDataKeys.indexOf(s.dataKey) > -1 )
					found = w.$EscortDeckSelectedPad;
				w.$EscortDeckSelectedPad++;
			}
			w.$EscortDeckSelectedPad = found;
			if( start && found == -1 )
				player.consoleMessage( "You have not any playable ships on board", 5 );
		}
*/
//		while( w.$EscortDeckSelectedPad < w.$EscortDeckPadPos.length
//			&& !w.$EscortDeckShip[w.$EscortDeckSelectedPad] )  //skip empty pads
//			w.$EscortDeckSelectedPad++; - no skip to be able to lock to a specific pad
		if( !w.$EscortDeckToWS || w.$EscortDeckSelectedPad != w.$EscortDeckPadPos.length ) {
		    if( w.$EscortDeckSelectedPad >= w.$EscortDeckPadPos.length ) {
			if( w.$EscortDeckSitIntoWithMode //allowing sit into enabled with mode key
				&& p.dataKey.indexOf("carrier-player") > -1 )//and player in a Carrier
				w.$EscortDeckSelectedPad = -2; //setup sit into mode
			else w.$EscortDeckSelectedPad = -1; //none selected
			var ws = w.$EscortDeckTWS; //Telescope worldScript
			if( ws ) ws.$TelescopeTargetSet = true;
			p.target = null;
			if( ws ) ws.$TelescopeTargetSet = false;
		    }
		    if( w.$EscortDeckSelectedPad > -1 ) {
			var s = w.$EscortDeckShip[w.$EscortDeckSelectedPad];
			if( s && s.isValid ) {
				var ws = w.$EscortDeckTWS; //Telescope worldScript
				if( ws ) {
					var mi = ws.$TelescopeList.indexOf( s );
					if( mi == -1 ) ws.$Telescope_Scan(); //forced rescan
					if( mi > -1 ) {
						if(  ws.$TelescopeListi != mi + 1 ) {
							ws.$TelescopeListi = mi + 1;
							ws.$Telescope_Show2( false ); //lock target
						}
					}
				} else p.target = s; //target the selected ship
			}
		    } else w.$EscortDeck_EQDisplayMode(w);
		}
		w.$EscortDeck_MFD(w); //update MFD
	}
}

this.$EscortDeck_FCB = function( delta ) { //FrameCallBack updating positions of escort ships
	var owner = player.ship; //modify this for NPCs
	if( !owner || !owner.isValid ) return; //exit if player died
	var w = worldScripts.escortdeck;
	var zv = w.$EscortDeckZV;

	var eship = w.$EscortDeckShip;
	var epos = w.$EscortDeckShipPos;
	if( owner != player.ship && owner.script && owner.script.$EscortDeckZV ) { //NPC
		zv = owner.script.$EscortDeckZV;
		eship = owner.script.$EscortDeckShip;
		epos = owner.script.$EscortDeckShipPos;
	} else if( w.$EscortDeckCWS && w.$EscortDeckCWS.$CarriersPlayerInEscort ) {
		owner = w.$EscortDeckCWS.$CarriersCarrier; //deck is no more on player ship but on Carrier
		if( !owner || !owner.isValid ) return; //exit if carrier died
	}
	if( !zv ) {
		if( !zv ) zv = Vector3D(0,0,-50);
		zv.x = 0;
		var hl = owner.boundingBox.z * 0.5; //half of the ship's length
		if( zv.z < hl ) zv.z = -hl; //correct z position for carriers
		if( zv.y > 0 ) zv.y = 0; //do not raise over the center line
		w.$EscortDeckZV = zv;
	}

	var d = w.$EscortDeckShipData;
	var sh = w.$EscortDeckSpawnedShips.pop();
	if( sh && sh.isValid ) { //restore ship data after ShipVersion changes in shipSpawned
		var pad = eship.indexOf( sh );
		if(w.$Debug) log(w.name, "FCB pad:"+pad+" d:"+d[pad]);
		if( pad > -1 && d[pad] )
			w.$EscortDeck_RestoreShipData(sh, d[pad], w, owner);
	}
//	if(w.$Debug) log(w.name, "FCB pad:1 d:"+d[1]);

	//set the new position of the deck if exist (carriers haven't added deck)
	var dp = owner.position.add(owner.heading.multiply(zv.z)); //deck position, used at escorts also
	if( zv.y != 0 ) dp = dp.add(owner.orientation.vectorUp().multiply(zv.y)); //move lower if needed
	if( w.$EscortDeckV ) { //set the new position of the deck if exist (carriers haven't added deck)
		w.$EscortDeckV.position = dp;
		w.$EscortDeckV.orientation = owner.orientation;
	}
	
	//set the new positions of escorts
	for(var pad = 0; pad < w.$EscortDeckPadPos.length; pad++ ) {
		var s = eship[pad];
		if( s ) {
			if( s.isValid ) {
				var sp = epos[pad]; //on deck and is not a Carrier?
				if( sp && s.dataKey.indexOf("carrier") == -1 ) {
					var h = owner.heading;
					var o = owner.orientation;
					var r = o.vectorRight();
					var u = o.vectorUp();
					var pos = dp.add(h.multiply(sp.z)).add(r.multiply(sp.x))
						.add(u.multiply(sp.y));
					if( pos.magnitude() > 500 ) s.position = pos;

					if( w.$EscortDeckCWS
						&& owner.dataKey.indexOf("carrier") > -1 ) {
						w.$EscortDeck_FCBC(s, owner, pad, o, h, r, u, w);
					} else if( w.$EscortDeckXL ) {
						var left = -0.5; //left side of the xl deck
						if( Math.round(pad/2) != pad/2 ) left = 0.5; //right
						var p = -Math.PI;
						s.orientation=o.rotate(r,p*0.5001).rotate(h,left*p);
					} else s.orientation = o.rotate( r, 0.0001 );
					//if(pad==6 || pad==7) reverse ori
				}
			} else { //escort ship destoyed
				if( w.$EscortDeckShip ) {
					if( owner != player.ship && ( !w.$EscortDeckCWS
						|| owner != w.$EscortDeckCWS.$CarriersCarrier ) )
						owner.script.$EscortDeckShip[pad] = null; //NPC escort
					else w.$EscortDeckShip[pad] = null; //player escort
				}
			}
		}
	}
	
	w.$EscortDeckFCBDelta += delta;
	if( w.$EscortDeckFCBDelta > 0.5 ) { //update AI and MFD twice in a second
		w.$EscortDeckFCBDelta = 0;
		//check nearby usable derelict for pickup
		var t = owner.target;
//		log(w.name, "target:"+t); //debug
		if(t) w.$EscortDeck_AddEscort( t, eship, owner, w, true );
		
		//AI helper part, keep in sync with the UPDATE parts of EscortDeck_AI.plist
		for(var pad = 0; pad < w.$EscortDeckPadPos.length; pad++ ) {
			var s = eship[pad];
			if( s && s.isValid && s.AI == "EscortDeck_AI.plist" ) { //skip waipoint AIs
				var e = s.script;
				if( s.AIState == "FOLLOW" ) e._locatePlayer2(s);
				else if( s.AIState == "GOTO" ) e._findPlayerHostiles2(s);
				else if( s.AIState == "ATTACK" || s.AIState == "INJECT"
					|| s.AIState == "SNIPER" ) e._combatCheck2(s);
				else if( s.AIState == "FLEE" ) e._fleeCheck2(s);
				else if( s.AIState == "LANDING" ) e._landing2(s);
			}
		}

		w.$EscortDeck_MFD(w); //update MFD
	}
}

this.$EscortDeck_FCBC = function(s, owner, pad, o, h, r, u, w) { //FrameCallBack rotate ships on Carriers
	var left = -0.5001; //left side of the xl deck
	if( Math.round(pad/2) != pad/2 ) left = 0.5001; //right side
	var p = -Math.PI;
	if( w.$EscortDeck_isMiniPad(pad, owner, w) ) {
		if( w.$EscortDeckXL ) left = -left;
		s.orientation = o.rotate( h, left*p ).rotate( u, left*p*0.6 ).rotate( r, 0.0001 );
	} else if( w.$EscortDeckXL && pad > 3 )
		s.orientation = o.rotate( u, left*p ).rotate( r, 0.0001 );
	else s.orientation = o.rotate( r, 0.0001 ); //fix blinking shaders bug
}

this.$EscortDeck_Interface = function() {
	var ps = player.ship; if( !ps || !ps.isValid || !ps.docked ) return;
	ps.dockedStation.setInterface(this.name,{
		title:"Escort Deck",
		category:"Ship Systems",
		summary:"Show and adjust your escort ships on Escort Deck.",
		callback:this.$EscortDeck_InterfaceCallBack.bind(this)
	});
}

this.$EscortDeck_InterfaceCallBack = function() {
	var ps = player.ship; if( !ps || !ps.isValid ) return;
	player.ship.hudHidden = true;//to shift down the menu
	var w = worldScripts.escortdeck;
	var d = w.$EscortDeckShipData;
	var dk = "escortdeck";
	var curChoices = {};
	var exitscr = "GUI_SCREEN_INTERFACES";
	var ex = "99_EXIT";
	var ic = ex;
	var mp = 0;
	var msg = "";
	var j = 0;
	var len = w.$EscortDeckPadPos.length;
	if( w.$EscortDeckSelectedPad < 0 ) w.$EscortDeckSelectedPad = 0;
	if( ps.equipmentStatus("EQ_ESCORTDECK") != "EQUIPMENT_OK"
		&& ps.equipmentStatus("EQ_ESCORTDECKXL") != "EQUIPMENT_OK" ) {
		msg = "You have no Escort Deck, buy it in the Equipment shop.";
		exitscr = "GUI_SCREEN_EQUIP_SHIP";
	} else for( var pad = 0; pad < len && j++ < 20; pad++ ) {
		var al = "LEFT";
		var un = false; //unselectable
		var co = "orangeColor";
		var s = d[pad];
		var tx = "Deck"+(pad+1)+" ";
		var x = 0, y = 0, z = 0, maxmass = 0;
		var si = w.$EscortDeckPadSize;
		if( si && si[0] ) {
			var e = null;
			var len = si.length-1;
			if( pad < 0 || pad > len ) e = si[len]; //the last line contain the maximums
			else e = si[pad];
			if( e && e[0] ) {
				x = e[0];
				y = e[1];
				z = e[2];
				maxmass = e[3];
			}
		}
		tx += "("+Math.floor(maxmass/1000)+"t, "+x+"x"+y+"x"+z+"m): ";
		if( s && s[0] && s[0][3] ) {
			tx += s[0][3].substring(0,38);// + " Fuel: "+Math.round(s[9]*10)/10+"ly";
//			msg += this.$EscortDeck_MFDAlign(tx, s.AIState)+"\n";
		} else {
			tx += "Empty";
			co = "grayColor";
		}
		tx += "\n";
		var cc = "01_" + ( pad < 10 ? "0" : "" ) + pad;
		curChoices[cc] = {text:tx, color:co, unselectable:un, alignment:al};
		if( pad == w.$EscortDeckSelectedPad ) {
			ic = cc; //initialChoicesKey
			msg = "Deck"+(pad+1)+": ";
			if( d[pad] && d[pad][1] ) {
				dk = d[pad][1];//dataKey
				mp = d[pad][13];//personality
				msg += this.EscortDeck_InterfaceShipData(d[pad]);
			} else msg += "Empty";
		}
	}

//	curChoices["20_DOIT"] = {text:"Do something", color:"yellowColor"};

	curChoices[ex] = {text:"Exit", color:"yellowColor"};
	var opts = {
		screenID: "escortdeck",
		title: "Escort Deck",
		allowInterrupt: false,
		exitScreen: exitscr,
		choices: curChoices,
		initialChoicesKey: ic,
		message: msg,
		model: "[" + dk + "]",
		modelPersonality: mp,
		spinModel:false
	};
	mission.runScreen(opts, this.$EscortDeck_InterfaceChoice, this);
	var m = mission.displayModel;
	if(m) m.orientation = m.orientation.rotateZ( -Math.PI/2 );
}


this.$EscortDeck_InterfaceChoice = function(choice) {
//	log(this.name, choice+" "+choice.indexOf("01_")+" "+choice.substr(3));
	if (choice == "99_EXIT") {
		player.ship.hudHidden = false;
		return;
	}
	if (choice.indexOf("01_") == 0) {
		this.$EscortDeckSelectedPad = 1*choice.substr(3);
		this.$EscortDeck_InterfaceCallBack(); //display the menu again
	} else if (choice == "20_DOIT" ) {
		//...
	}
}

this.EscortDeck_InterfaceShipData = function(d, all) {
	if(!d || !d[1]) return("No data");
	var psd = d[0][3]+"\n";//ship.displayName+
	var equip="";
	var s="";
	s=d[3]; if(s) s=EquipmentInfo.infoForKey(s);//ship.forwardWeapon;
	if( s!=null && s.equipmentKey != "EQ_WEAPON_NONE" ) equip+="Forward "+s.name;
	s=d[2]; if(s) s=EquipmentInfo.infoForKey(s)//ship.aftWeapon;
	if( s!=null && s.equipmentKey != "EQ_WEAPON_NONE" ) {
		if( equip.length > 0 ) equip+=", "; //Anaconda has aft weapon only
		equip+="Aft "+s.name;
	}
	s=d[4]; if(s) s=EquipmentInfo.infoForKey(s)//ship.portWeapon;
	if( s!=null && s.equipmentKey != "EQ_WEAPON_NONE" ) equip+=", Port "+s.name;
	s=d[5]; if(s) s=EquipmentInfo.infoForKey(s)//ship.starboardWeapon;
	if( s!=null && s.equipmentKey != "EQ_WEAPON_NONE" ) equip+=", Starboard "+s.name;
	if(equip.length > 0) equip+="\n";
	else equip+="No Laser\n";
	var eqs = "";
	var savedEQ = d[7], counter;
	for(counter = 0; counter < savedEQ.length; counter++) {
		if( savedEQ[counter][1] != "EQUIPMENT_DAMAGED" ) {
//			eqs += savedEQ[counter][0] + ", ";
 			s = EquipmentInfo.infoForKey(savedEQ[counter][0]);
			if( s ) {
				var n = s.name;
				if( n.length > 0 ) eqs += n + ", ";
			}
		}
	}
	eqs = eqs.substr( 0, eqs.length - 2 ); //cut last comma
	if( eqs.length > 0 ) equip+=eqs+"\n";
	eqs = "";
	for(counter = 0; counter < savedEQ.length; counter++) {
		if( savedEQ[counter][1] == "EQUIPMENT_DAMAGED" ) {
 			s = EquipmentInfo.infoForKey(savedEQ[counter][0]);
			if( s ) {
				var n = s.name;
				if( n.length > 0 ) eqs += n + ", ";
			}
		}
	}
	eqs = eqs.substr( 0, eqs.length - 2 ); //cut last comma
	if( eqs.length > 0 ) equip+="Damaged: "+eqs+"\n";

	var desc = "";
/*	
	if( ship.scriptInfo && ship.scriptInfo.buydesc )//show ship description if any
		desc="\""+ship.scriptInfo.buydesc+"\"\n";
	var sh = 0;
	if( ship == player.ship ) sh += ship.maxForwardShield;
	else {
		var w = worldScripts["NPC-shields"];
		if( w && ship.script ) {
			if( !ship.script.maxShieldStrength ) w.shipSpawned(ship); //add npc shield
			sh += ship.script.maxShieldStrength;
		}
		var w = worldScripts["customshields"];
		if( w && ship.script ) {
			if( !ship.script.customshieldsmaxforwardshieldlevel ) w.shipSpawned(ship); //add customshield
			sh += ship.script.customshieldsmaxforwardshieldlevel;
		}
	}
	if( ship.dataKey.indexOf("escape-capsule") > -1 ) sh = 0; //no shield on escape pod
	else if( worldScripts["shieldequalizercapacitors"] ) {
		if( ship.equipmentStatus("EQ_BIGSHCAP") == "EQUIPMENT_OK"
			&& ship.equipmentStatus("EQ_FORWARD_SHIELD_CAPACITOR") == "EQUIPMENT_OK" ) sh += 256;
		else if( ship.equipmentStatus("EQ_FORWARD_SHIELD_CAPACITOR") == "EQUIPMENT_OK" ) sh += 64;
	}
	var hd = "";
	if(!ship.hasHyperspaceMotor ) hd = "  No Hyperdrive";
	var speed = ship.maxSpeed;
	if( ship.script && ship.script.name === "rocketships" )
		speed = Math.round(ship.maxSpeed/3*2)+"+"+Math.round(ship.maxSpeed/3);
	if( speed > 0 ) psd += "Speed: "+speed+" mLS  Thrust: "+ship.maxThrust+"  ";
	if( all || speed > 0 && ship.isPiloted ) {
		psd += "\nPitch: "+Math.round(ship.maxPitch*100)/100+
			"  Roll: "+Math.round(ship.maxRoll*100)/100+
			"  Yaw: "+Math.round(ship.maxYaw*100)/100+hd+
//			"  Version: "+this.$shipVersion(ship)+
//			"  Service "+ship.serviceLevel+
			"\nCargo: "+ship.cargoSpaceCapacity;
		if( ship.extraCargo > 0 ) psd += "+"+parseInt(ship.extraCargo); //need Oolite v1.79
		else if( ship.scriptInfo && ship.scriptInfo.cargoext && parseInt(ship.scriptInfo.cargoext) > 0 )
			psd += "+"+parseInt(ship.scriptInfo.cargoext);
		psd += "t  ";
	}
	var sp = "";
	var sph = Math.round(ship.collisionRadius*ship.collisionRadius*Math.PI);
	if( sph < 1000000 ) sp = sph;
	else sp = Math.round(sph/100000)/10+"k";//base sphere too large to fit in the line so show in km^2

	var en = "";
	var r = "";
	if(all) en = "Energy: "+ship.maxEnergy;
	else {
		if( ship.isPiloted ) {
			var eb = Math.max(1, Math.floor(ship.maxEnergy/64));
			en = "EBank";
			if( eb > 1 ) en += "s";
			en += ": "+eb;
		}
	}

	if( ship.isPiloted || all ) {
		var er = 0;
		if( ship.energyRechargeRate > 0 ) er = parseInt(ship.energyRechargeRate); //need Oolite v1.79
		else if( ship.scriptInfo && ship.scriptInfo.recharge && parseInt(ship.scriptInfo.recharge) > 0 )
			er = parseInt(ship.scriptInfo.recharge);
		if( er > 0 ) {
			var ers = "";
			if( er < 2.5 ) ers = "Poor";
			else if( er < 3.5 ) ers = "Medium";
			else if( er < 4.5 ) ers = "Good";
			else if( er < 10 ) ers = "Excellent";
			else ers = "Extreme";
			if( all ) ers += " ("+er+")";
			r = "  Recharge: "+ers;
		}
		if( sh > 0 ) r += "\nShields: "+Math.max(2, Math.floor(sh/64));
		if( ship.missileCapacity > 0 ) {
			if( sh > 0 ) r += "  "; else r += "\n";
			r += "Missile";
			if( ship.missileCapacity > 1 ) r += "s";
			r += ": "+ship.missileCapacity;
		}
	}
*/
//{usable:u, mass:ship.mass, x:b.x, y:b.y, z:b.z});//storedShipArray[14]
	if( d[14] ) psd += "Mass: "+Math.max(Math.round(d[14].mass/100)/10, 1)+ //min.1t
		"t  Size: "+Math.round(d[14].x)+"*"+
		Math.round(d[14].y)+"*"+Math.round(d[14].z)+"m";
//		"m Radius: "+Math.round(ship.collisionRadius)+"m"+//" Sphere: "+sp+"m^2"+
//		"\n"+en+r;
	psd += "  Fuel: "+Math.round(d[9]*10)/10+" ly\n"+equip;
	psd = desc + psd;
//	if( ship.price > 0 ) psd += "\nPrice: "+formatCredits(ship.price, false, true);
	return(psd);
}


this.$EscortDeck_isFreePadFor = function(eqKey, owner, context, w, mini) { 
	//called from escortship equipment conditions
	//should check TooLarge but no ship nor mass,etc. data available
	//due to shouldn't spawn ships in equipment conditions
	//so must provide in the prefilled $EscortDeckEqs array for TooLarge3
	
	var len = w.$EscortDeckPadPos.length;
	for(var i = 0; i < len; i++ ) {
		if( !w.$EscortDeckShipData[i] && //find the first empty pad where fit
			( !w.$EscortDeck_isMiniPad(i, owner, w) || mini ) ) {
			if( !w.$EscortDeckEqs || !w.$EscortDeckEqs[eqKey] ||
				!w.$EscortDeck_TooLarge3( owner, w,
					w.$EscortDeckEqs[eqKey][3],
					w.$EscortDeckEqs[eqKey][0],
					w.$EscortDeckEqs[eqKey][1],
					w.$EscortDeckEqs[eqKey][2], i ) )
				return true;
		}
	}
	return false;
}

this.$EscortDeck_isLarge = function(ship, w) { //check if the ship is over the large border
	var m = ship.mass; //revert HardShips' double mass caused by density=2
	if( ship.scriptInfo && ship.scriptInfo.hardarmour > 1 ) m = m / 2;
	if( m >= w.$EscortDeckLargeMass ) return true;
	return false;
}

this.$EscortDeck_isMini = function(ship) { //check if the ship is like an Adder or smaller
	if( !ship || !ship.isValid ) return false;
	var b = ship.boundingBox;
	if( !b || ship.mass > this.$EscortDeckPad.mini[3]
		|| b.x > this.$EscortDeckPad.mini[0]
		|| b.y > this.$EscortDeckPad.mini[1]
		|| b.z > this.$EscortDeckPad.mini[2] ) return false;
	return true;
}

this.$EscortDeck_isMiniPad = function(pad, owner, w) { //check if this pad is for mini ships only
	if( w.$EscortDeckCWS && owner.dataKey.indexOf("carrier") > -1 ) {
		if( w.$EscortDeckXL ) {
			if( pad != 8 && pad != 9 ) return false;
		} else if( pad != 7 && pad != 8 ) return false;
	} else if( w.$EscortDeckXL || pad != 4 && pad != 5 ) return false;
	return true;
}

this.$EscortDeck_isPadFree = function(pad, owner, w) { //is this pad free for this ship?
	var s = null; //or the assigned ship is not valid or not in scanner range
	if( w && w.$EscortDeckShip ) s = w.$EscortDeckShip[pad];
	if( w && pad < w.$EscortDeckPadPos.length && ( !s || !s.isValid 
//		|| owner && owner.isValid && owner.position.distanceTo(s.position) > owner.scannerRange 
		|| pad == w.$EscortDeckSelectedPad ) ) { //pad is selected to replace ship?
		return(true);
	}
	return(false);
}

this.$EscortDeck_MakeShipData = function(ship) {
	if( !ship || !ship.isValid ) return null;
	var storedShipArray = new Array; //like in ShipStorageHelper
	var stationAttributes;
	if (ship.dockedStation) {
		stationAttributes = [ship.dockedStation.primaryRole,
			ship.dockedStation.dataKey, ship.dockedStation.position];
	} else stationAttributes = [null,null,null];
	storedShipArray.push([galaxyNumber, system.ID, clock.seconds, ship.displayName,
			     system.name, stationAttributes, ship.price]); //storedShipArray[0]
	storedShipArray.push(ship.dataKey); // storedShipArray[1]
	if (!ship.aftWeapon){storedShipArray.push(ship.aftWeapon);}
	else {storedShipArray.push(ship.aftWeapon.equipmentKey);} // storedShipArray[2]
	if (!ship.forwardWeapon){storedShipArray.push(ship.forwardWeapon);}
	else {storedShipArray.push(ship.forwardWeapon.equipmentKey);} // storedShipArray[3]
	if (!ship.portWeapon){storedShipArray.push(ship.portWeapon);}
	else {storedShipArray.push(ship.portWeapon.equipmentKey);} // storedShipArray[4]
	if (!ship.starboardWeapon){storedShipArray.push(ship.starboardWeapon);}
	else {storedShipArray.push(ship.starboardWeapon.equipmentKey);} // storedShipArray[5]
	var missiles = [];
	if( ship.missileCapacity > 0 ) { //bugfix of "Tried to init array with nil object" in Oolite 1.81
		missiles = ship.missiles;
		var counter;
		for(counter = 0;counter<missiles.length;counter++)
			missiles.splice(counter,1,missiles[counter].equipmentKey);
	}
	storedShipArray.push(missiles); // storedShipArray[6]
	var equipment = ship.equipment;
	var tempArray;
	for(counter = 0;counter<equipment.length;counter++) {
		tempArray = [equipment[counter].equipmentKey, ship.equipmentStatus(equipment[counter])];
		equipment.splice(counter,1,tempArray);
	}
	storedShipArray.push(equipment); // storedShipArray[7]
	storedShipArray.push(ship.cargoList); // storedShipArray[8]
	storedShipArray.push(ship.fuel); //storedShipArray[9]
	storedShipArray.push(new Array); //passengers is player only, storedShipArray[10]
	var subEnts = new Array;
	if (ship.subEntityCapacity > 0 && ship.subEntityCapacity !== ship.subEntities.length) {
		// only store subEnts if at less than capacity
		for (counter = 0;counter < ship.subEntities.length; counter++) {
			subEnts.push([ship.subEntities[counter].dataKey,
				ship.subEntities[counter].position.x,
				ship.subEntities[counter].position.y,
				ship.subEntities[counter].position.z]);
		}
	}
	storedShipArray.push(subEnts);//storedShipArray[11]

	if( ship == player.ship ) {
		storedShipArray.push(ship.serviceLevel);//storedShipArray[12]
		ship.script.$SSH_serviceLevel = ship.serviceLevel; //save for Carriers OXP
	} else if( ship.script ) storedShipArray.push(ship.script.$SSH_serviceLevel);
	
	storedShipArray.push(ship.entityPersonality);//storedShipArray[13]
	
	var u = 0; //escortdeck specific data
	if( ship.script && ship.script.$EscortDeckUsable > 0 )
		u = ship.script.$EscortDeckUsable;
	var b = ship.boundingBox; //mass for TooLarge and ship price, sizes needed in playerboughtnewship
	storedShipArray.push({usable:u, mass:ship.mass, x:b.x, y:b.y, z:b.z});//storedShipArray[14]
	
	return(storedShipArray);
}

this.$EscortDeck_MassFactor = function(ship, w) { //compute thrust reduction due to extra mass in deck
	var m = 0; //total mass of ships on deck
	for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
		var d = w.$EscortDeckShipData[i]; //inside stations must use the saved data array
		var s = w.$EscortDeckShip[i];
		if( ( d || s ) && w.$EscortDeckShipPos[i] ) { //there is a ship in this deck
			if( d && d[14] && d[14].mass > 0 ) m += d[14].mass;
			else if( s && s.mass > 0 ) m += s.mass;
			else m += 50000; //guess
		}
	}
	//in proportion with the original mass (max. 1)
	return( ship.mass / ( ship.mass + m / 3 ) );
}

this.$EscortDeck_MFD = function(w) { //update MFD
	var ps = player.ship; if( !ps || !ps.isValid ) return;
	var d = w.$EscortDeckShipData;
	var mfd = "";
	var j = 0;
	var len = w.$EscortDeckPadPos.length;
	if( w.$EscortDeckDockingTunnel ) ; //show an empty MFD while the player travel in the tunnel
	else if( ps.equipmentStatus("EQ_ESCORTDECK") != "EQUIPMENT_OK"
		&& ps.equipmentStatus("EQ_ESCORTDECKXL") != "EQUIPMENT_OK" )
		mfd = "No Escort Deck\n";
	else for( var pad = 0; pad < len && j++ < 10; pad++ ) {
		var s = w.$EscortDeckShip[pad];
		var state = "";
		if( w.$EscortDeckLocked[pad] ) state = "LOCKED";
		var l = (pad+1)+": ";
		if( pad == w.$EscortDeckSelectedPad ) l += "-> "; //mark selected pad
		if( s && s.isValid ) {
			if( state.length == 0 ) state = s.AIState; //only if not locked
			if( s.script && s.script.$EscortDeckUsable > 0  ) { //show for usables only
				l += s.name.substring(0,18)+" E"+Math.round(s.energy/64);
					//+"/"+Math.round(s.maxEnergy/64);
					//first 18 char to fit into but show Sidewinder Special
				//l += " E"+Math.round(s.energy);
//				var r = "";
//				var t = Math.round(s.maxEnergy/64); //total letters
//				for(var k = 0; k < t; k++ )
//					if( k < Math.round(s.energy/64) ) r += "E";
//					else r += "-"; //empty bank
//				l += " "+r;
				if( s.equipmentStatus("EQ_FUEL_INJECTION") == "EQUIPMENT_OK" )
					l += " F"+Math.ceil(s.fuel-0.01);
				if( !w.$EscortDeckShipPos[pad] ) //not in deck
					l += " "+Math.floor(ps.position.distanceTo(s.position)/1000)+"km "
						+w.$EscortDeck_MFDFrom(s.position);
			} else l += s.displayName+" ("+Math.round(s.mass/1000)+"t)";
			mfd += this.$EscortDeck_MFDAlign(l, state)+"\n";
		} else {
			var x = 0, y = 0, z = 0, maxmass = 0;
			var s = w.$EscortDeckPadSize;
			if( s && s[0] ) {
				var e = null;
				var len = s.length-1;
				if( pad < 0 || pad > len ) e = s[len]; //the last line contain the maximums
				else e = s[pad];
				if( e && e[0] ) {
					x = e[0];
					y = e[1];
					z = e[2];
					maxmass = e[3];
				}
			}
			l += "("+Math.floor(maxmass/1000)+"t, "+x+"x"+y+"x"+z+"m)";
			if( state.length > 0 ) mfd += this.$EscortDeck_MFDAlign(l, state)+"\n";
			else mfd += l+"\n";
		}
	}

	var t = this.$EscortDeckToWS; //Towbar
	if( t && j < 10 ) {
		var l = "T: ";
		if( j == w.$EscortDeckSelectedPad ) l += "-> "; //mark selected pad
		var s = t.$TowbarShip;
		if( s ) {
			l += s.displayName+" ("+Math.round(s.mass/1000)+"t)";
			mfd += this.$EscortDeck_MFDAlign(l, "SALVAGE")+"\n";
		} else mfd += l+"("+Math.floor(ps.mass*1.6/1000)+"t)\n"; //max. mass on towbar
	}
		
	if( ps.setMultiFunctionText ) ps.setMultiFunctionText(w.name, mfd, false);
}

this.$EscortDeck_MFDAlign = function(left, right) { //insert as many spaces between as fill the line
	var width = 14.2;
	var space = " ";
	left += space;
	var t = left + right;
	while( defaultFont.measureString(t) < width ) {
		left += space;
		t = left + right;
	}
	//fine tune with hair spaces - from spara's marketObserver
	var hairSpace = String.fromCharCode(31);
	while ( defaultFont.measureString(t + hairSpace) < width ) {
		left += hairSpace;
		t = left + right;
	}
	return(t);
}

this.$EscortDeck_MFDFrom = function(whomposition) { //direction mark
	var ship = player.ship;
	if( !ship || !ship.isValid || !whomposition || !whomposition.dot ) return ("");
	var v = ship.position.subtract(whomposition);
	var fw = ship.vectorForward.angleTo(v);
	var ri = ship.vectorRight.angleTo(v);
	var up = ship.vectorUp.angleTo(v);
	var s = ""; // refine if out of front 1 degree cone
	if( ri < 1.56 ) s += "<";
	if( up > 1.58 ) s += "^";
	else if( up < 1.56 ) s += "v";
	if( ri > 1.58 ) s += ">";
//	if( s.length > 0 || fw < 1 ) //do not exclude the aft 1 degree cone
//		s = Math.round( 180 * ( 1 - fw / Math.PI ) ) + "° " + s;
	return( s );
}


this.$EscortDeck_PutShip = function(ship, pad, owner, w, msg, land, xch ) { //put ship into the pad
	var r = w.$EscortDeck_TooLarge2(ship, owner, w, pad);
	if( r ) return( r ); //too large
	if( ship && ship.isValid && owner && owner.isValid
		&& ( land || w.$EscortDeck_isPadFree(pad, owner, w) ) ) { //landing back or new ship
		if( !w.$EscortDeckShip ) w.$EscortDeckShip = [];
		var old = w.$EscortDeckShip[pad];
		var sp = w.$EscortDeckShip.indexOf(ship); //ship pad if already owned
		if( old && old.isValid && !xch ) { //prevent recursion during exchange with old ship
			if( sp > -1 && sp != pad ) { //put an owned ship to other pad
				//exchange with the old ship if old fit into the another pad
				//if( w.$EscortDeck_PutShip(old, sp, owner, w, msg, true, true ) )
				w.$EscortDeck_Release(owner, pad, w, //relock old if fit
						      w.$EscortDeckLandingDist-100, msg);
				w.$EscortDeckSelectedPad = -1; //prevent lock to the new pad
				owner.target = old; //for relock
			} else if( ship != old )
				w.$EscortDeck_Release(owner, pad, w, false, msg); //replace old ship
		}
		if( sp > -1 ) { //remove from the previous pad
			var d = [], s = [];
			for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
				if( i == sp ) {
					s[i] = null;
					d[i] = null;
				} else {
					s[i] = w.$EscortDeckShip[i];
					d[i] = w.$EscortDeckShipData[i];
				}
			}
			w.$EscortDeckShipData = d;
			w.$EscortDeckShip = s;
		}
		w.$EscortDeckShip[pad] = ship;
//		if( !w.$EscortDeckShipData[pad] || land ) - was a try to exclude first time after load game but not helped
			w.$EscortDeckShipData[pad] = w.$EscortDeck_MakeShipData(ship);
		if(w.$Debug) log(w.name, (pad+1)+". "+w.$EscortDeckShipData[pad]); //debug
		var usable = 0;
		if( ship.script && ship.script.$EscortDeckUsable > 0 )
			usable = ship.script.$EscortDeckUsable;
		
		if( usable > 0 && ship.isDerelict ) { //removing derelict flag need respawn
			var ship2 = w.$EscortDeck_SpawnShip( pad, w, owner );
			if( ship2 && ship2.isValid ) {
				ship.remove(true);
				w.$EscortDeckShip[pad] = ship = ship2;
			} //else fallback to the derelict ship
		}

		if( ship.AI != "EscortDeck_AI.plist" ) ship.switchAI("EscortDeck_AI.plist");
		var js = "escortdeck_escort";
		if( !ship.script || ship.script.name != js ) {
			var sl = 0;  //must set usable and sl again after setScript
			if( ship.script ) sl = ship.script.$SSH_serviceLevel;
			ship.setScript(js+".js");
			if( ship.script ) {
				ship.script.$EscortDeckUsable = usable;
				ship.script.$SSH_serviceLevel = sl;
			}
		}
			
		if( usable > 0 ) {
			ship.AIState = "DECK";
			if( !owner.group ) { //escort group for performEscort AI command
				owner.group = new ShipGroup();
				owner.group.addShip(owner);
				owner.group.leader = owner;
			}
			if( owner && owner.group.leader != owner ) owner.group.leader = owner;
			if( !owner.group.containsShip(ship) ) owner.group.addShip(ship);
		} else ship.AIState = "SALVAGE";
		ship.beaconCode = "E"+pad+" "+ship.name;
		ship.bounty = 0;
		ship.scanClass = "CLASS_CARGO"; //prevent masslock
		ship.scannerDisplayColor1 = [0,0,0]; //black
// 
		var carrier = false;
		if( w.$EscortDeckCWS && owner.dataKey.indexOf("carrier") > -1 ) carrier = true;
		var p = w.$EscortDeckPadPos[pad];
		if( !p ) {
			w.$EscortDeck_SetPads( owner, w );
			p = w.$EscortDeckPadPos[pad];
			if( !p ) p = [0,70*(pad+1),0];
		}
		var sp = Vector3D(p[0], p[1], p[2]);
		var x = ship.boundingBox.x * 0.5;
		var y = ship.boundingBox.y * 0.5;
		var z = ship.boundingBox.z * 0.5;
		if( w.$EscortDeckXL && ( !carrier || pad > 3 && pad < 8 ) ) {
			var left = -0.5; //left side of the xl deck
			if( Math.round(pad/2) != pad/2 ) left = 0.5; //right side
			sp = sp.add(Vector3D( left * 2 * z, 0, 0));
		} else {
			var v = w.$EscortDeckPadVec[pad];
			if( !v ) v = [0,1,0];//up
			var vx = v[0] * x;
			if( !carrier && y > 12 ) y = 12 - ( y - 12 ); //sunk if taller than 24m
			var vy = v[1] * y;
			var vz = v[2] * z;
			sp = sp.add(Vector3D(vx, vy, vz));
		}
		if(owner.addCollisionException) owner.addCollisionException(ship); //from v1.81
		else if( carrier ) {  //must move up to avoid kill by Injector speed
			var a = y + 25;
			if( w.$EscortDeck_isMiniPad( pad, owner, w) ) a = 2 * x + 30;
			sp = sp.add(Vector3D(0, a, 0));
		}

		if( sp.magnitude() > 500 ) ship.position = sp; //prevent terminated by Witchpoint Beacon
		if( owner != player.ship && ship.script ) {
			if(!ship.script.$EscortDeckShipPos) 
				ship.script.$EscortDeckShipPos = [];
			ship.script.$EscortDeckShipPos[pad] = sp; //NPC
		} else w.$EscortDeckShipPos[pad] = sp; //player ship
		ship.orientation = owner.orientation;
		ship.velocity = owner.velocity;
		
		//slower steering and acceleration in proportion with the mass on deck
		w.$EscortDeck_SetMaxs( owner, w );
		
		if( ship.script ) {
			ship.script.$Towbar_TimedBombCounter = -1; //disarm timed bomb
			if( ship.script.$TowbarUsableShip )
				ship.script.$TowbarUsableShip = null;
		}
		
		if( owner == player.ship ) {
			if( msg ) {
				var m = ship.displayName+" locked on deck "+(pad+1);
				player.consoleMessage(m, 10);
				if(w.$Debug) log(w.name, m); //debug
				w.$EscortDeckSound.play();
			}
		}
		
		return false;
	}
	return("no room");
}

this.$EscortDeck_Release = function(owner, pad, w, dist, nomsg) {
	if( !dist ) var dist = w.$EscortDeckLandingDist + 500; //drop distance over lock range
	var s = w.$EscortDeckShip[pad];
	if( s && s.isValid ) {
		//move the old ship backward, over the relock range if on deck
		if( w.$EscortDeckShipPos[pad] ) {
			var sp = s.position.add(player.ship.heading.multiply(-dist));
			if( sp.magnitude() > 500 ) s.position = sp; //prevent terminated by Witchpoint Beacon
			w.$EscortDeckSound.play();
		}
		if( owner.removeCollisionException ) //from v1.81
			owner.removeCollisionException(s);
		w.$EscortDeckShipData[pad] = null;
		w.$EscortDeckShip[pad] = null;
		w.$EscortDeckShipPos[pad] = null;
		if(!nomsg && s.dataKey.indexOf("carrier-player") == -1 ) //no msg for carrier
			player.consoleMessage( s.displayName
			+" released from deck "+(pad+1), 10 );
	}
}

this.$EscortDeck_RestoreShipData = function(s, d, w, owner) {
	if( s && s.isValid && d && d[0] ) {
		if( !s.script ) s.setScript("oolite-default-ship-script.js"); //for sure

		if( s != player.ship ) s.displayName = d[0][3];
		if(d[2]) s.aftWeapon = EquipmentInfo.infoForKey(d[2]);
		if(d[3]) s.forwardWeapon = EquipmentInfo.infoForKey(d[3]);
		if(d[4]) s.portWeapon = EquipmentInfo.infoForKey(d[4]);
		if(d[5]) s.starboardWeapon = EquipmentInfo.infoForKey(d[5]);

		var counter, counter1;
		var equipment = s.equipment;
		//remove all equipments
		var r = [];
		for(counter = 0;counter<equipment.length;counter++)
			r.push(equipment[counter].equipmentKey);
		for(counter = 0;counter<r.length;counter++) //separated to avoid problems during eq array shortening
			s.removeEquipment(r[counter]);
//		if(w.$Debug) log(w.name, "RestoreShipData2 d:"+d);
//		if(w.$Debug) log(w.name, "RestoreShipData2e "+equipment);
		var tempArray = [];
		var savedEQ = d[7];
		for(counter = 0; counter < savedEQ.length; counter++) {
			tempArray.push(savedEQ[counter][0]);
		}
		for(counter = 0; counter < equipment.length; counter++) {
			var e = equipment[counter].equipmentKey;
			var index = tempArray.indexOf(e);
			if( index == -1 ) s.removeEquipment(e);
			else if( savedEQ[index][1] == "EQUIPMENT_DAMAGED" )
				s.setEquipmentStatus( e, "EQUIPMENT_DAMAGED" );
		}
		var depth = 0;
		var retryArray = []; //try again if a required equipment comes later
		retryArray[depth] = savedEQ;
		do {
		    var added = false;
		    retryArray[depth+1] = [];
		    for(counter = 0; counter < retryArray[depth].length; counter++) {
			var e = retryArray[depth][counter][0]; //award if exists and has not
			var ed = retryArray[depth][counter][1];
			if( EquipmentInfo.infoForKey(e) ) {
				if( !s.awardEquipment(e) ) //retry next time if failed
					retryArray[depth+1].push([e,ed]);
				else {
					added = true;
					if( ed == "EQUIPMENT_DAMAGED" )
						s.setEquipmentStatus( e, "EQUIPMENT_DAMAGED" );
				}
			}
		    }
		    depth++;
//		    if(w.$Debug) log(w.name, "RestoreShipData3 depth:"+depth+" retryArray:"+retryArray[depth]);
		} while( retryArray[depth].length > 0 && added && depth < 100 ) //until no more added eq
		/* old code lose equipments due to make changes in the original d array
		for(counter = 0; counter<equipment.length; counter++) {
			tempArray = [equipment[counter].equipmentKey,
				s.equipmentStatus(equipment[counter])];
			for(counter1 = 0; counter1<savedEQ.length; counter1++) {
				if (savedEQ[counter1][0] === tempArray[0]) {
					tempArray.push(true);
					tempArray.push(counter1);
					if (savedEQ[counter1][1] === "EQUIPMENT_DAMAGED")
						s.setEquipmentStatus(tempArray[0],
								     "EQUIPMENT_DAMAGED");
					break;
				}
			}
			if (!tempArray[2]) {
				s.removeEquipment(tempArray[0]);
				continue;
			} else savedEQ.splice(tempArray[3],1);
		}
		for(counter = 0; counter<savedEQ.length; counter++) {
			if(EquipmentInfo.infoForKey(savedEQ[counter][0])) {
				s.awardEquipment(savedEQ[counter][0]);
				if(savedEQ[counter][1] === "EQUIPMENT_DAMAGED")
					s.setEquipmentStatus(savedEQ[counter][0],savedEQ[counter][1]);
			}
		}*/
//		if(w.$Debug) log(w.name, "RestoreShipData3 d[7]:"+d[7]);
//		if(w.$Debug) log(w.name, "RestoreShipData3e "+s.equipment);

		//ship.cargoList is read-only and no ship.manifest so d[8] is skipped
		s.fuel = d[9];

		//passengers is player only, so d[10] is skipped
		var subEnts = d[11]; //will only exist if less subEnts than capacity when ship was stored.
		if (subEnts && subEnts.length>0) {
			var currentSubEnts = s.subEntities;
			for (counter = 0; counter<currentSubEnts.length; counter++) {
				currentSubEnts[counter].toBeRemoved = true;
		    		for(counter1 = 0; counter1<subEnts.length; counter1++) {
				if(subEnts[counter1][0] === currentSubEnts[counter].dataKey &&
				 subEnts[counter1][1] === currentSubEnts[counter].position.x &&
				 subEnts[counter1][2] === currentSubEnts[counter].position.y &&
				 subEnts[counter1][3] === currentSubEnts[counter].position.z )
					delete currentSubEnts[counter].toBeRemoved;
				}
				if (currentSubEnts[counter].toBeRemoved) {
					delete currentSubEnts[counter].toBeRemoved;
					currentSubEnts[counter].remove(true);
				}
			}
		}

		//serviceLevel is player only, NPCs store it in script variable
		if( s == player.ship && d[12] ) s.serviceLevel = d[12];
		else if( s.script ) s.script.$SSH_serviceLevel = d[12]; //must set after setScript

		if (oolite.compareVersion("1.80") < 0) {// can only do this in Oolite 1.81 or greater
			s.entityPersonality = d[13];
		}
		
		if( s.script && d[14] )
			s.script.$EscortDeckUsable = d[14].usable; //must set after setScript
			
//		if(w.$Debug) log(w.name, "missileCapacity:"+s.missileCapacity+" d[6]:"+d[6]);
		if( s.missileCapacity > 0 ) {
		//do missiles at the end due to there is a core bug in Oolite 1.80
		//when land back to the carrier from an Asp which has max_missiles=0:
		//Error: Native exception: Tried to init array with nil object
			var missiles = [];
			if( s.missiles && s.missiles.length > 0 ) {
				missiles = s.missiles;
				for(counter = 0;counter<missiles.length;counter++){
					s.removeEquipment(missiles[counter]);}
			}
			missiles = d[6];
			if( missiles && missiles.length > 0 ) {	
				for(counter = 0; counter<missiles.length; counter++)
					if(EquipmentInfo.infoForKey(missiles[counter]))
						s.awardEquipment(missiles[counter]);
			}
		}
	}
	return( s );
}

this.$EscortDeck_RollBack = function(eq, msg) {
	var f = 1;
	if( player.ship && player.ship.dockedStation ) {
		var e = player.ship.dockedStation.equipmentPriceFactor;
		if( e > 0 ) f = e;
	}
	player.credits += f * EquipmentInfo.infoForKey(eq).price/10;
	player.consoleMessage(msg, 5);
}

this.$EscortDeck_Salvage = function( pad, w, owner ) { //salvage the escort on the given pad
	var d = null;
	if( w.$EscortDeckShipData ) d = w.$EscortDeckShipData[pad];
	var s = null;
	if( w.$EscortDeckShip ) s = w.$EscortDeckShip[pad];
	var sn = "escort";
	var t = w.$EscortDeckToWS;
	if( t ) {
		var s = w.$EscortDeck_SpawnShip( pad, w, owner );
		if( s && s.isValid ) t.$Towbar_Payout(s); //call payout in towbar oxp
	} else { //pay salvage price like in towbar
		var cr = w.$EscortDeck_SalvagePadPrice( pad, w ); //salvage payment
		log(w.name, "Salvage "+pad+". pad for "+cr+"cr: "+d);
		if( s && s.isValid ) sn = s.displayName;
		else {
			if( !d || !d[1] ) return false; //no ship in this pad
			sn = d[0]; //displayName
		}
		player.consoleMessage("You got "+cr+" credits for salvaging "+sn, 10);
		player.credits += cr;
		this.$EscortDeckSound2.play();
	}
	if( s && s.isValid ) s.remove(true);
	w.$EscortDeckShipData[pad] = null;
	w.$EscortDeckShip[pad] = null;
}

this.$EscortDeck_SalvageBadEscorts = function(owner, w) {
	if( !owner ) return;
	w.$EscortDeck_SetPads( owner, w );
	if( owner.mass < w.$EscortDeckMinMass ) { //bought a too small ship
		owner.removeEquipment("EQ_ESCORTDECK");
		owner.removeEquipment("EQ_ESCORTDECKXL");
		for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) //salvage all escorts
			w.$EscortDeck_Salvage(i, w, owner);
	} else if( !w.$EscortDeck_isLarge(owner, w) ) { //no side pads
		w.$EscortDeck_Salvage(4, w, owner); //salvage left adder
		w.$EscortDeck_Salvage(5, w, owner); //salvage right adder
	}
	if( owner.mass < w.$EscortDeckXLMass ) { //bought a non-xl ship
		if( owner.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" ) {
			owner.removeEquipment("EQ_ESCORTDECKXL"); //downgrade
			owner.awardEquipment("EQ_ESCORTDECK");
		}
	}
	if( owner.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" ) w.$EscortDeckXL = true;
	else w.$EscortDeckXL = false; //must set XL for the following TooLarge call
	for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
		var t = null;
		if( w.$EscortDeckShip ) t = w.$EscortDeckShip[i]; //the ship on this pad
		if( this.$EscortDeck_TooLarge2(t, owner, w, i) ) //can use ShipData array if no ship
			w.$EscortDeck_Salvage(i, w, owner); //salvage too large escorts
	}
}

this.$EscortDeck_SalvagePadPrice = function( pad, w ) { //salvage price of an escort on this pad
	var s = null;
	if( w.$EscortDeckShip ) s = w.$EscortDeckShip[pad]; //the ship on this pad if player is not docked
	if( s && s.isValid ) return( w.$EscortDeck_SalvageShipPrice( s, w ) );
	var d = null;
	if( w.$EscortDeckShipData ) d = w.$EscortDeckShipData[pad]; //inside stations must use the saved data array
	if( !d || !d[1] ) return( 0 ); //no price for a nonexistent ship
	var eq = [];
	for( var i = 0; i < d[7].length; i++) {
		var ep = 0;
		var e = EquipmentInfo.infoForKey(d[7][i][0]);
		if(e) ep = e.price;
		eq[i] = {price:ep};
		if(w.$Debug) log(w.name, i+". "+e+" fullprice:"+eq[i].price); //debug
	}
	var mass = 30000;//guess
	if( d[14] && d[14].mass > 0 ) mass = d[14].mass;
	return( w.$EscortDeck_SalvagePrice( mass, eq, d[3], d[2], w ) );
}

this.$EscortDeck_SalvagePrice = function( mass, eq, fw, af, w ) {//price like in towbar OXP
	var hull = Math.round( 32.8 * Math.min(80, Math.round( //1t alloys = 32.8cr in average
		0.3 * mass * 0.001 )) * 0.1 ) * 10 //alloys from the hull

	var comp = 0; if( mass > 30000 ) comp = 84; //price of 1t computers

	var eqcr = 0; //payment for all equipments
	var rnd = Math.random();
	var i = -1;
	while( eq[++i] ) {
		var p = eq[i].price * 0.01; //max. 10% of the original cost in credits (not in credits*10)
		if( p > 300 ) {
			rnd = rnd * ( 300 / p ); //costly eqs always damaged
			eqcr += Math.round( Math.max( 50,
				Math.round(p * rnd ))/10)*10; //at least 50cr
		} else eqcr += Math.round( Math.max( 10,
			Math.round( p * rnd ))/10)*10; //at least 10cr
	}
	
	//payment for fw and aft weapons, min. 10%, max 20% of original price
	var fc = 0;
	if( fw ) fc = Math.round((EquipmentInfo.infoForKey(fw).price/10*(0.1+rnd*0.1))/10)*10;
	var ac = 0;
	if( af ) ac = Math.round((EquipmentInfo.infoForKey(af).price/10*(0.1+rnd*0.1))/10)*10;
		
	var cr = hull + comp + eqcr + fc + ac; //all payments but nothing for cargo
	return( cr * 0.5 ); //deduct 50% commission as salvaging without license
}

this.$EscortDeck_SalvageShipPrice = function( ship, w ) { //salvage price of this ship
	return( w.$EscortDeck_SalvagePrice( ship.mass, ship.equipment,
					  ship.forwardWeapon, ship.aftWeapon, w ) );
}

this.$EscortDeck_SaveMaxs = function( ship, w ) {
	if(ship) w.$EscortDeckMaxs = { maxThrust:ship.maxThrust,
		maxPitch:ship.maxPitch,
		maxRoll:ship.maxRoll,
		maxYaw:ship.maxYaw };
	else w.$EscortDeckMaxs = { maxThrust:32, maxPitch:1, maxRoll:2, maxYaw:1 };
}

this.$EscortDeck_SetMaxs = function( ship, w ) {
	if( !ship || !ship.addCollisionException  //need Oolite v1.81
		|| !w.$EscortDeckMassEffect ) return;
	var f = w.$EscortDeck_MassFactor(ship, w);
	var m = w.$EscortDeckMaxs; //original maximums of player ship
	ship.thrust = f * ship.maxThrust;
	ship.maxPitch = f * m.maxPitch;
	ship.maxRoll = f * m.maxRoll;
	ship.maxYaw = f * m.maxYaw;
}

this.$EscortDeck_SetPads = function( owner, w ) {
	var carrier = false;
	if( w.$EscortDeckCWS && owner.dataKey.indexOf("carrier") > -1 ) carrier = true;
	var p = w.$EscortDeckLPos;
	var v = w.$EscortDeckLVec;
	var xl = "escortdeck";
	var m, s = [], si, x, y, z, maxmass;
	if( owner && owner.isValid ) {
		if( owner.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" ) {
			w.$EscortDeckXL = true;
		} else w.$EscortDeckXL = false;

		if( carrier ) {
			xl = null; //use the carrier's hull as deck
			if( w.$EscortDeckXL ) { //XL deck on carrier
				p = w.$EscortDeckCXPos;
				v = w.$EscortDeckCXVec;
				si = w.$EscortDeckCXSize;
			} else { //normal deck on carrier
				p = w.$EscortDeckCPos;
				v = w.$EscortDeckCVec;
				si = w.$EscortDeckCSize;
			}
			for( var i = 0; i < p.length; i++ ) {
				var siz = si[i];
				if( siz ) {
					x = siz[0];
					y = siz[1];
					z = siz[2];
					maxmass = siz[3];
				} else x = y = z = maxmass = 0;
				s[i] = [x, y, z, maxmass];
				if( i == 0 ) m = [x, y, z, maxmass];
				else {
					m[0] = Math.max(x, m[0]);
					m[1] = Math.max(y, m[1]);
					m[2] = Math.max(z, m[2]);
					m[3] = Math.max(maxmass, m[3]);
				}
			}
			s.push(m); //the last item contain the maximums of size values
		} else if( w.$EscortDeckXL ) { //XL deck on non-carrier
			xl += "xl";
			p = w.$EscortDeckXLPos;
			v = w.$EscortDeckXLVec;
			x = w.$EscortDeckPad.xl[0];
			y = w.$EscortDeckPad.xl[1];
			z = w.$EscortDeckPad.xl[2];
			maxmass =  w.$EscortDeckPad.xl[3]; //60t
			for( var i = 0; i <= p.length; i++ ) { // <= due to all equal
				s[i] = [x, y, z, maxmass]; //so the maximums are the same
			}
		} else { //normal deck on non-carrier
			if( owner.mass >= w.$EscortDeckWMass ) {
				xl += "w";
				p = w.$EscortDeckWPos;
				v = w.$EscortDeckWVec;
			}
			if( owner.mass >= w.$EscortDeckWMass ) { //wide pad
				x = w.$EscortDeckPad.wide[0];
				y = w.$EscortDeckPad.wide[1];
				z = w.$EscortDeckPad.wide[2];
				maxmass = w.$EscortDeckPad.wide[3];
			} else { //normal escort pad (non-wide)
				x = w.$EscortDeckPad.escort[0];
				y = w.$EscortDeckPad.escort[1];
				z = w.$EscortDeckPad.escort[2];
				maxmass = w.$EscortDeckPad.escort[3];
			}
			maxmass = Math.min(owner.mass * 0.4,  maxmass); //max. 40% of the owner's mass
			maxmass = Math.min(w.$EscortDeckLargeMass,  maxmass); //max. 130t
			for( var i = 0; i <= p.length; i++ ) {
				if( w.$EscortDeck_isMiniPad(i, owner, w) )
					 s[i] = [w.$EscortDeckPad.mini[0],
					 	w.$EscortDeckPad.mini[1],
						w.$EscortDeckPad.mini[2],
						w.$EscortDeckPad.mini[3]];
				else s[i] = [x, y, z, maxmass];
			}
		}
	}

	var len = p.length;
	if( owner && owner.isValid && !w.$EscortDeck_isLarge(owner, w) )
		len = 4; //without adder pads
	w.$EscortDeckPadPos = []; //landing pad positions
	w.$EscortDeckPadSize = []; //landing pad sizes
	w.$EscortDeckPadVec = []; //vectors from landing pad to ship center
	for(var i = 0; i < len; i++ ) { //fill up from the proper array
		w.$EscortDeckPadPos[i] = p[i];
		w.$EscortDeckPadSize[i] = s[i];
		w.$EscortDeckPadVec[i] = v[i];
	}
	w.$EscortDeckPadSize[len] = s[s.length-1]; //the last item contain the maximums of size values
	return( xl );
}

this.$EscortDeck_Show = function(owner) { //create the deck and escorts
	var w = worldScripts.escortdeck;
	var xl = "notcarrier"; //carriers set xl to null
	if( ( owner.equipmentStatus("EQ_ESCORTDECK") == "EQUIPMENT_OK" 
		|| owner.equipmentStatus("EQ_ESCORTDECKXL") == "EQUIPMENT_OK" )
		&& owner.mass >= w.$EscortDeckMinMass ) {
		xl = w.$EscortDeck_SetPads( owner, w ); //return null for Carriers
		if( !w.$EscortDeckV && xl ) {
			var r = system.addVisualEffect( xl, [0,0,0] );
			if( r ) w.$EscortDeckV = r; //position and orientation will be set in FCB
		}
		
		//make the z offset of deck position from the center of the owner ship
		var zv = owner.weaponPositionAft;//to be able to fire backward if the deck is solid
		if(!zv.z) zv = zv[0]; //fix for the change of weaponPositions in Oolite 1.83.
		zv.x = 0;
		var hl = owner.boundingBox.z * 0.5; //half of the ship's length
		if( zv.z < hl ) zv.z = -hl; //correct z position for carriers
		if( zv.y > 0 ) zv.y = 0; //do not raise over the center line
		if( owner == player.ship ) w.$EscortDeckZV = zv;
		else { //NPC deck
			if( !owner.script ) owner.script = "oolite-default-ship-script.js"; //for sure
			owner.script.$EscortDeckZV = zv;
		}

		//spawn ships based on $EscortDeckShipData
		var b = system.shipsWithPrimaryRole("buoy-witchpoint"); //avoid instant collision after hyperjump
		for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) { //fill up ShipPos array
			var s = w.$EscortDeck_SpawnShip(i, w, owner);
			if( s && s.isValid ) {
				if( b && b[0] && b[0].addCollisionException) //from v1.81
					b[0].addCollisionException(s);
				var r = w.$EscortDeck_PutShip(s, i, owner, w);
				if(w.$Debug) log(w.name, (i+1)+". spawn "+s+" r:"+r); //debug
			}
		}
	}
	w.$EscortDeck_MFD(w); //update MFD

	if( ( w.$EscortDeckV || !xl) && !isValidFrameCallback( w.$EscortDeckFCB ) )
		w.$EscortDeckFCB = addFrameCallback( w.$EscortDeck_FCB );
}

this.$EscortDeck_SpawnOne = function( role, w ) {
	var s = system.addShips( role, 1, [10000000,0,0], 5000 );
	return( s );
}

this.$EscortDeck_SpawnShip = function( pad, w, owner ) {
	var s = null;
	var d = w.$EscortDeckShipData[pad];
	if( d && d[1] ) {
		if( w.$EscortDeckCWS && d[14] && d[14].usable > 0
			&& w.$EscortDeckPlayableDataKeys.indexOf(d[1]) == -1 ) {
			var n = d[1]+"-player"; //try the player variant for Carriers
			if( w.$EscortDeckPlayableDataKeys.indexOf(n) > -1 ) d[1] = n;
			else {
				var e = d[1].indexOf("-trader"); //try without suffix
				if( e == d[1].length-(e.length) ) {
					n = d[1].substr(0, e);
					if( w.$EscortDeckPlayableDataKeys.indexOf(n) > -1 )
						d[1] = n;
					else {
						n += "-player"; //try replace -trader to -player
						if( w.$EscortDeckPlayableDataKeys.indexOf(n) > -1 )
							d[1] = n;
					}
				} else {
					var e = d[1].indexOf("-pirate"); //try without suffix
					if( e == d[1].length-(e.length) ) {
						n = d[1].substr(0, e);
						if( w.$EscortDeckPlayableDataKeys.indexOf(n) > -1 )
							d[1] = n;
						else {
							n += "-player"; //try replace to -player
							if( w.$EscortDeckPlayableDataKeys
								.indexOf(n) > -1  )d[1] = n;
						}
					}
				}
			}
		}
		s = w.$EscortDeck_SpawnOne( "["+d[1]+"]", w );
	}
	if(w.$Debug) log(w.name, "Spawn escort pad:"+pad+" d:"+d+" s:"+s); //debug
	if( s && s[0] ) {
		s = w.$EscortDeck_RestoreShipData(s[0], d, w, owner);
		//setup restore in FCB also, need after ShipVersion changes in shipSpawned
		w.$EscortDeckSpawnedShips.push(s);
	}
	return( s );
}

this.$EscortDeck_TooLarge = function(ship, owner, w) { //size or mass is too large to this deck?
	return(w.$EscortDeck_TooLarge2(ship, owner, w, -1));
}

this.$EscortDeck_TooLarge2 = function(ship, owner, w, pad) { //use saved data if no ship
	if( !ship || !ship.isValid ) {
		if( pad < 0 ) return("invalid escort ship");
		var d = w.$EscortDeckShipData;
		if( !d ) return("missing escort ship data");
		d = d[pad];
		if( !d || !d[14] || !d[14].mass ) return("invalid escort ship data");
		var shipmass = d[14].mass;
		if( d[1].indexOf("hard") == 0 ) //if this is a HardShip or Granite ship
			shipmass = shipmass / 2; //remove the extra density
		var bx = d[14].x;
		var by = d[14].y;
		var bz = d[14].z;
	} else { //ship is present
		var shipmass = ship.mass;
		if( ship.scriptInfo && ship.scriptInfo.hardarmour > 1 )
			shipmass = shipmass / 2; //remove extra density from HardShips and Granite ships
		var b = ship.boundingBox;
		var bx = b.x;
		var by = b.y;
		var bz = b.z;
	}
	return(w.$EscortDeck_TooLarge3(owner, w, shipmass, bx, by, bz, pad ));
}

this.$EscortDeck_TooLarge3 = function(owner, w, shipmass, bx, by, bz, pad ) { //check given data
	if( !owner || !owner.isValid )  return("invalid owner ship");
	if( !( shipmass > 0 ) ) return("missing escort ship mass");
	if( !bx || !(bx>0) || !by || !(by>0) || !bz || !(bz>0) )
		return("invalid escort ship size");
	var e, l, x = 0, y = 0, z = 0, maxmass = 0;
	var s = w.$EscortDeckPadSize;
	if( s && s[0] ) {
		l = s.length-1;
		if( pad < 0 || pad > l ) e = s[l]; //the last line contain the maximums
		else e = s[pad];
		if( e && e[0] ) { 
			x = e[0];
			y = e[1];
			z = e[2];
			maxmass = e[3];
		}
	}

	var r = null;
	if( shipmass > maxmass ) r = "too heavy ("+Math.ceil(shipmass/1000)
		+"t > "+Math.floor(maxmass/1000)+"t)";
	else if( bx > x ) r = "too wide ("+Math.ceil(bx)+"m > "+x+"m)";
	else if( by > y ) r = "too tall ("+Math.ceil(by)+"m > "+y+"m)";
	else if( bz > z ) r = "too long ("+Math.ceil(bz)+"m > "+z+"m)";
	if(w.$Debug) {
		var p = pad+1;
		if( p == 0 ) p = "any";
		if( r ) var l = r;
		else var l = "["+Math.ceil(bx*100)/100+","+
			Math.ceil(by*100)/100+","+
			Math.ceil(bz*100)/100+","+Math.ceil(shipmass)
			+"] fit into ["+Math.floor(x*100)/100+","+
			Math.floor(y*100)/100+","+
			Math.floor(z*100)/100+","+Math.floor(maxmass)+"]";
		log(w.name, "Pad "+p+": "+l); //debug
	}
	return( r ); //if null then not too large
}

this.$EscortDeck_VClear = function(w) { //remove the models of deck and escorts
	var p = player.ship;
	if( isValidFrameCallback( w.$EscortDeckFCB ) )
		removeFrameCallback( w.$EscortDeckFCB );
	if( w.$EscortDeckV ) w.$EscortDeckV.remove();
	w.$EscortDeckV = null;
	w.$EscortDeckSelectedPad = -1;
	if( w.$EscortDeckShip ) {
		for(var i = 0; i < w.$EscortDeckPadPos.length; i++ ) {
			var s = w.$EscortDeckShip[i];
			if( s && s.isValid ) {
				var lost = false;
				if( !w.$EscortDeckShipPos[i] ) { //not in deck
					var rnd = ( 5 * Math.random() + 1 ) * p.scannerRange;
					var d = s.position.distanceTo(p.position) - p.scannerRange;
					if( d > rnd ) lost = true; //leaved farther than its luck
				}
				if( lost ) player.consoleMessage(s.name
					+" is left behind from pad "+i, 10);
				else {
					w.$EscortDeckShipData[i] = w.$EscortDeck_MakeShipData(s);
					s.remove(true);
				}
				if(w.$Debug) log(w.name, (i+1)+". "+w.$EscortDeckShipData[i]); //debug
			}
			w.$EscortDeckShip[i] = null;
			w.$EscortDeckShipPos[i] = null;
		}
	} else {
		w.$EscortDeckShip = [];
		w.$EscortDeckShipPos = [];
	}
}